# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Common utility functions and data structures."""

from __future__ import annotations

import gc
import sys
from abc import ABC, abstractmethod
from collections import OrderedDict
from collections.abc import Mapping, Set
from concurrent.futures import Future
from contextlib import contextmanager
from enum import Enum
from functools import cached_property, partial, wraps
from operator import (
	add,
	attrgetter,
	eq,
	floordiv,
	ge,
	gt,
	le,
	lt,
	mod,
	mul,
	pow,
	sub,
	truediv,
)
from random import Random
from textwrap import dedent
from time import monotonic
from types import FunctionType, MethodType, ModuleType
from typing import (
	TYPE_CHECKING,
	Annotated,
	Callable,
	Hashable,
	Iterable,
	Iterator,
	KeysView,
	MutableMapping,
	Optional,
	Sequence,
	Type,
	TypeVar,
)

import msgpack
import networkx as nx
import numpy as np
from annotated_types import Ge, Le
from blinker import Signal
from tblib import Traceback

from . import exc
from .exc import TimeError, WorkerProcessReadOnlyError
from .graph import DiGraph, Edge, GraphMapping, Node
from .types import (
	Branch,
	CharName,
	EdgeValDict,
	EternalKey,
	NodeName,
	NodeValDict,
	RulebookName,
	RuleName,
	Stat,
	Tick,
	Time,
	TimeWindow,
	Turn,
	UniversalKey,
	Value,
)
from .wrap import SpecialMapping

if TYPE_CHECKING:
	from .rule import Rule, RuleBook
	from .xcollections import FunctionStore

TRUE: bytes = msgpack.packb(True)
FALSE: bytes = msgpack.packb(False)
NONE: bytes = msgpack.packb(None)
EMPTY: bytes = msgpack.packb({})
ELLIPSIS: bytes = b"\xc7\x00{"
NAME: bytes = msgpack.packb("name")
NODES: bytes = msgpack.packb("nodes")
EDGES: bytes = msgpack.packb("edges")
UNITS: bytes = msgpack.packb("units")
RULEBOOK: bytes = msgpack.packb("rulebook")
RULEBOOKS: bytes = msgpack.packb("rulebooks")
NODE_VAL: bytes = msgpack.packb("node_val")
EDGE_VAL: bytes = msgpack.packb("edge_val")
ETERNAL: bytes = msgpack.packb("eternal")
UNIVERSAL: bytes = msgpack.packb("universal")
STRINGS: bytes = msgpack.packb("strings")
RULES: bytes = msgpack.packb("rules")
TRIGGERS: bytes = msgpack.packb("triggers")
PREREQS: bytes = msgpack.packb("prereqs")
ACTIONS: bytes = msgpack.packb("actions")
NEIGHBORHOOD: bytes = msgpack.packb("neighborhood")
BIG: bytes = msgpack.packb("big")
LOCATION: bytes = msgpack.packb("location")
BRANCH: bytes = msgpack.packb("branch")
EMPTY_MAPPING: bytes = msgpack.packb({})


class SignalDict(Signal, dict):
	def __setitem__(self, __key, __value):
		super().__setitem__(__key, __value)
		self.send(self, key=__key, value=__value)

	def __delitem__(self, __key):
		super().__delitem__(__key)
		self.send(self, key=__key, value=None)


class BadTimeException(Exception):
	"""You tried to do something that would make sense at a different game-time

	But doesn't make sense now

	"""


class MsgpackExtensionType(Enum):
	"""Type codes for packing special lisien types into msgpack"""

	tuple = 0x00
	frozenset = 0x01
	set = 0x02
	exception = 0x03
	graph = 0x04
	character = 0x7F
	place = 0x7E
	thing = 0x7D
	portal = 0x7C
	ellipsis = 0x7B
	function = 0x7A
	method = 0x79
	trigger = 0x78
	prereq = 0x77
	action = 0x76


class get_rando:
	"""Attribute getter for randomization functions

	Aliases functions of a randomizer, wrapped so that they won't run in
	planning mode, and will save the randomizer's state after every call.

	"""

	__slots__ = ("_getter", "_wrapfun", "_instance")
	_getter: callable

	def __init__(self, attr, *attrs):
		self._getter = attrgetter(attr, *attrs)

	def __get__(self, instance, owner) -> callable:
		if hasattr(self, "_wrapfun") and self._instance is instance:
			return self._wrapfun
		retfun: callable = self._getter(instance)

		@wraps(retfun)
		def remembering_rando_state(*args, **kwargs):
			if instance._planning:
				raise exc.PlanError("Don't use randomization in a plan")
			ret = retfun(*args, **kwargs)
			instance.universal["rando_state"] = instance._rando.getstate()
			return ret

		self._wrapfun = remembering_rando_state
		self._instance = instance
		return remembering_rando_state


@contextmanager
def timer(msg="", logfun: callable = None):
	if logfun is None:
		logfun = print
	start = monotonic()
	yield
	logfun("{:,.3f} {}".format(monotonic() - start, msg))


def getatt(attribute_name):
	"""An easy way to make an alias"""
	return property(attrgetter(attribute_name))


_T = TypeVar("_T")


def singleton_get(s: Iterable[_T]) -> _T:
	"""Take an iterable and return its only item if possible, else None."""
	it = None
	for that in s:
		if it is not None:
			return None
		it = that
	return it


class EntityAccessor(ABC):
	__slots__ = (
		"engine",
		"entity",
		"branch",
		"turn",
		"tick",
		"stat",
		"current",
		"mungers",
	)

	def __init__(
		self,
		entity: GraphMapping | Node | Edge,
		stat: Stat,
		engine: AbstractEngine | None = None,
		branch: Branch | None = None,
		turn: Turn | None = None,
		tick: Tick | None = None,
		current: bool = False,
		mungers: list[Callable] | None = None,
	):
		if engine is None:
			engine = entity.db
		if branch is None:
			branch = engine.branch
		if turn is None:
			turn = engine.turn
		if mungers is None:
			mungers = []
		self.current = current
		self.engine = engine
		self.entity = entity
		self.stat = stat
		self.branch = branch
		self.turn = turn
		self.tick = tick
		self.mungers = mungers

	def __ne__(self, other):
		return self() != other

	def __str__(self):
		return str(self())

	def __repr__(self):
		return "EntityStatAccessor({}[{}]{}), {} mungers".format(
			self.entity,
			self.stat,
			""
			if self.current
			else ", branch={}, turn={}, tick={}".format(
				self.branch, self.turn, self.tick
			),
			len(self.mungers),
		)

	def __gt__(self, other):
		return self() > other

	def __ge__(self, other):
		return self() >= other

	def __lt__(self, other):
		return self() < other

	def __le__(self, other):
		return self() <= other

	def __eq__(self, other):
		return self() == other

	def munge(self, munger: callable):
		return EntityStatAccessor(
			self.entity,
			self.stat,
			self.engine,
			self.branch,
			self.turn,
			self.tick,
			self.current,
			self.mungers + [munger],
		)

	def __add__(self, other):
		return self.munge(partial(add, other))

	def __sub__(self, other):
		return self.munge(partial(sub, other))

	def __mul__(self, other):
		return self.munge(partial(mul, other))

	def __rpow__(self, other, modulo=None):
		return self.munge(partial(pow, other, modulo=modulo))

	def __rdiv__(self, other):
		return self.munge(partial(truediv, other))

	def __rfloordiv__(self, other):
		return self.munge(partial(floordiv, other))

	def __rmod__(self, other):
		return self.munge(partial(mod, other))

	def __contains__(self, item):
		return item in self()

	def __getitem__(self, k):
		return self.munge(lambda x: x[k])

	@abstractmethod
	def _get_value_now(self) -> Value: ...

	def __call__(
		self,
		branch: Branch | None = None,
		turn: Turn | None = None,
		tick: Tick | None = None,
	):
		if self.current:
			res = self._get_value_now()
		else:
			time_was = self.engine.time
			self.engine.branch = branch or self.branch
			self.engine.turn = turn if turn is not None else self.turn
			if tick is not None:
				self.engine.tick = tick
			elif self.tick is not None:
				self.engine.tick = self.tick
			res = self._get_value_now()
			self.engine.time = time_was
		for munger in self.mungers:
			res = munger(res)
		return res

	def iter_history(self, beginning: Turn, end: Turn) -> Iterator[Value]:
		"""Iterate over all the values this stat has had in the given window, inclusive."""
		# It might be useful to do this in a way that doesn't change the
		# engine's time, perhaps for thread safety
		engine = self.engine
		oldturn = engine.turn
		oldtick = engine.tick
		for turn in range(beginning, end + 1):
			engine.turn = turn
			try:
				y = self._get_value_now()
			except KeyError:
				yield None
				continue
			if hasattr(y, "unwrap"):
				y = y.unwrap()
			yield y
		engine.turn = oldturn
		engine.tick = oldtick


class UnitsAccessor(EntityAccessor):
	entity: AbstractCharacter

	def _get_value_now(self) -> dict[CharName, list[NodeName]]:
		ret = {}
		for graph in self.entity.unit:
			ret[graph] = []
			for node in self.entity.unit[graph]:
				ret[graph].append(node)
		return ret


class CharacterStatAccessor(EntityAccessor):
	entity: AbstractCharacter

	def _get_value_now(self) -> Value:
		return self.entity.stat[self.stat]


class EntityStatAccessor(EntityAccessor):
	def _get_value_now(self) -> Value:
		return self.entity[self.stat]


def dedent_source(source):
	nlidx = source.index("\n")
	if nlidx is None:
		raise ValueError("Invalid source")
	while source[:nlidx].strip().startswith("@"):
		source = source[nlidx + 1 :]
		nlidx = source.index("\n")
	return dedent(source)


def _sort_set_key(v):
	if isinstance(v, tuple):
		return (2,) + tuple(map(repr, v))
	if isinstance(v, str):
		return 1, v
	return 0, repr(v)


class SizedDict(OrderedDict):
	"""A dictionary that discards old entries when it gets too big."""

	def __init__(self, max_entries: Annotated[int, Ge(0)] = 1000):
		self._n = max_entries
		super().__init__()

	def __setitem__(self, key, value):
		while len(self) > self._n:
			self.popitem(last=False)
		super().__setitem__(key, value)


_sort_set_memo = SizedDict()


def sort_set(s: Set[_T]) -> list[_T]:
	"""Return a sorted list of the contents of a set

	This is intended to be used to iterate over world state.

	Non-strings come before strings and then tuples. Tuples compare
	element-wise as normal.

	This is memoized.

	"""
	if not isinstance(s, Set):
		raise TypeError("sets only")
	s = frozenset(s)
	if s not in _sort_set_memo:
		_sort_set_memo[s] = sorted(s, key=_sort_set_key)
	return _sort_set_memo[s].copy()


class FakeFuture(Future):
	"""A 'Future' that calls its function immediately and sets the result"""

	def __init__(self, func: callable, *args, **kwargs):
		super().__init__()
		self.set_result(func(*args, **kwargs))


def fake_submit(func: callable, *args, **kwargs) -> FakeFuture:
	"""A replacement for `concurrent.futures.Executor.submit` that works in serial

	In testing, you may use, eg.,
	``@patch.object(executor, 'submit', new=fake_submit)``
	to make normally parallel operations serial.

	"""
	return FakeFuture(func, *args, **kwargs)


class AbstractBookmarkMapping(MutableMapping, Callable):
	@abstractmethod
	def __call__(self, key: Key) -> None: ...


class AbstractEngine(ABC):
	"""Parent class to the real Engine as well as EngineProxy.

	Implements serialization and the __getattr__ for stored methods.

	By default, the deserializers will refuse to create lisien entities.
	If you want them to, use my ``loading`` property to open a ``with``
	block, in which deserialized entities will be created as needed.

	"""

	thing_cls: type
	place_cls: type
	portal_cls: type
	char_cls: type
	character: Mapping[CharName, Type[char_cls]]
	eternal: MutableMapping[EternalKey, Value]
	universal: MutableMapping[UniversalKey, Value]
	rulebook: MutableMapping[RulebookName, "RuleBook"]
	rule: MutableMapping[RuleName, "Rule"]
	main_branch: Branch
	branch: Branch
	turn: Turn
	tick: Tick
	time: Time
	function: ModuleType | "FunctionStore"
	method: ModuleType | "FunctionStore"
	trigger: ModuleType | "FunctionStore"
	prereq: ModuleType | "FunctionStore"
	action: ModuleType | "FunctionStore"
	bookmark: AbstractBookmarkMapping
	_rando: Random
	_branches_d: dict[
		Optional[Branch], tuple[Optional[Branch], Turn, Tick, Turn, Tick]
	]

	@cached_property
	def logger(self):
		if hasattr(self, "_logger"):
			return self._logger
		from logging import getLogger

		return getLogger("lisien")

	def log(self, level, msg, *args, **kwargs):
		self.logger.log(level, msg, *args, **kwargs)

	def debug(self, msg, *args, **kwargs):
		self.log(10, msg, *args, **kwargs)

	def info(self, msg, *args, **kwargs):
		self.log(20, msg, *args, **kwargs)

	def warning(self, msg, *args, **kwargs):
		self.log(30, msg, *args, **kwargs)

	def error(self, msg, *args, **kwargs):
		self.log(40, msg, *args, **kwargs)

	def critical(self, msg, *args, **kwargs):
		self.log(50, msg, *args, **kwargs)

	def is_ancestor_of(self, parent: Branch, child: Branch) -> bool:
		"""Return whether ``child`` is a branch descended from ``parent``

		At any remove.

		"""
		branches = self.branches()
		if parent not in branches:
			raise ValueError("Not a branch", parent)
		if child not in branches:
			raise ValueError("Not a branch", child)
		if parent is None or parent == child or parent == self.main_branch:
			return True
		if child == self.main_branch:
			return False
		if self.branch_parent(child) == parent:
			return True
		return self.is_ancestor_of(parent, self.branch_parent(child))

	@cached_property
	def pack(self) -> Callable[[Value], bytes]:
		try:
			from lise_ormsgpack import packb

			return packb
		except ImportError:
			pass
		handlers = {
			type(...): lambda _: msgpack.ExtType(
				MsgpackExtensionType.ellipsis.value, b""
			),
			nx.Graph: lambda graf: msgpack.ExtType(
				MsgpackExtensionType.graph.value,
				packer(
					[
						"Graph",
						graf._node,
						graf._adj,
						graf.graph,
					]
				),
			),
			nx.DiGraph: lambda graf: msgpack.ExtType(
				MsgpackExtensionType.graph.value,
				packer(["DiGraph", graf._node, graf._adj, graf.graph]),
			),
			nx.MultiGraph: lambda graf: msgpack.ExtType(
				MsgpackExtensionType.graph.value,
				packer(["MultiGraph", graf._node, graf._adj, graf.graph]),
			),
			nx.MultiDiGraph: lambda graf: msgpack.ExtType(
				MsgpackExtensionType.graph.value,
				packer(["MultiDiGraph", graf._node, graf._adj, graf.graph]),
			),
			tuple: lambda tup: msgpack.ExtType(
				MsgpackExtensionType.tuple.value, packer(list(tup))
			),
			frozenset: lambda frozs: msgpack.ExtType(
				MsgpackExtensionType.frozenset.value, packer(list(frozs))
			),
			set: lambda s: msgpack.ExtType(
				MsgpackExtensionType.set.value, packer(list(s))
			),
			FunctionType: lambda func: msgpack.ExtType(
				getattr(MsgpackExtensionType, func.__module__).value,
				packer(func.__name__),
			),
			MethodType: lambda meth: msgpack.ExtType(
				MsgpackExtensionType.method.value, packer(meth.__name__)
			),
			Exception: lambda exc: msgpack.ExtType(
				MsgpackExtensionType.exception.value,
				packer(
					[
						exc.__class__.__name__,
						Traceback(exc.__traceback__).to_dict()
						if hasattr(exc, "__traceback__")
						else None,
					]
					+ list(exc.args)
				),
			),
		}

		def pack_handler(obj):
			if isinstance(obj, Exception):
				typ = Exception
			else:
				typ = type(obj)
			if typ in handlers:
				return handlers[typ](obj)
			elif isinstance(obj, DiGraph):
				return msgpack.ExtType(
					MsgpackExtensionType.character.value, packer(obj.name)
				)
			elif isinstance(obj, AbstractThing):
				return msgpack.ExtType(
					MsgpackExtensionType.thing.value,
					packer([obj.character.name, obj.name]),
				)
			elif isinstance(obj, Node):
				return msgpack.ExtType(
					MsgpackExtensionType.place.value,
					packer([obj.graph.name, obj.name]),
				)
			elif isinstance(obj, Edge):
				return msgpack.ExtType(
					MsgpackExtensionType.portal.value,
					packer(
						[
							obj.graph.name,
							obj.orig,
							obj.dest,
						]
					),
				)
			elif isinstance(obj, Mapping):
				return dict(obj)
			elif isinstance(obj, list):
				return list(obj)
			raise TypeError("Can't pack {}".format(typ))

		packer = partial(
			msgpack.packb,
			default=pack_handler,
			strict_types=True,
			use_bin_type=True,
		)
		return packer

	@cached_property
	def unpack(
		self,
	) -> Callable[
		[bytes],
		Value
		| nx.DiGraph
		| char_cls
		| place_cls
		| thing_cls
		| portal_cls
		| Exception
		| callable,
	]:
		char_cls = self.char_cls
		place_cls = self.place_cls
		portal_cls = self.portal_cls
		thing_cls = self.thing_cls
		excs = {
			# builtin exceptions
			"AssertionError": AssertionError,
			"AttributeError": AttributeError,
			"EOFError": EOFError,
			"FloatingPointError": FloatingPointError,
			"GeneratorExit": GeneratorExit,
			"ImportError": ImportError,
			"IndexError": IndexError,
			"KeyError": KeyError,
			"KeyboardInterrupt": KeyboardInterrupt,
			"MemoryError": MemoryError,
			"NameError": NameError,
			"NotImplementedError": NotImplementedError,
			"OSError": OSError,
			"OverflowError": OverflowError,
			"RecursionError": RecursionError,
			"ReferenceError": ReferenceError,
			"RuntimeError": RuntimeError,
			"StopIteration": StopIteration,
			"IndentationError": IndentationError,
			"TabError": TabError,
			"SystemError": SystemError,
			"SystemExit": SystemExit,
			"TypeError": TypeError,
			"UnboundLocalError": UnboundLocalError,
			"UnicodeError": UnicodeError,
			"UnicodeEncodeError": UnicodeEncodeError,
			"UnicodeDecodeError": UnicodeDecodeError,
			"UnicodeTranslateError": UnicodeTranslateError,
			"ValueError": ValueError,
			"ZeroDivisionError": ZeroDivisionError,
			# networkx exceptions
			"HasACycle": nx.exception.HasACycle,
			"NodeNotFound": nx.exception.NodeNotFound,
			"PowerIterationFailedConvergence": nx.exception.PowerIterationFailedConvergence,
			"ExceededMaxIterations": nx.exception.ExceededMaxIterations,
			"AmbiguousSolution": nx.exception.AmbiguousSolution,
			"NetworkXAlgorithmError": nx.exception.NetworkXAlgorithmError,
			"NetworkXException": nx.exception.NetworkXException,
			"NetworkXError": nx.exception.NetworkXError,
			"NetworkXNoCycle": nx.exception.NetworkXNoCycle,
			"NetworkXNoPath": nx.exception.NetworkXNoPath,
			"NetworkXNotImplemented": nx.exception.NetworkXNotImplemented,
			"NetworkXPointlessConcept": nx.exception.NetworkXPointlessConcept,
			"NetworkXUnbounded": nx.exception.NetworkXUnbounded,
			"NetworkXUnfeasible": nx.exception.NetworkXUnfeasible,
			# lisien exceptions
			"NonUniqueError": exc.NonUniqueError,
			"AmbiguousUserError": exc.AmbiguousLeaderError,
			"RulesEngineError": exc.RulesEngineError,
			"RuleError": exc.RuleError,
			"RedundantRuleError": exc.RedundantRuleError,
			"UserFunctionError": exc.UserFunctionError,
			"WorldIntegrityError": exc.WorldIntegrityError,
			"CacheError": exc.CacheError,
			"TravelException": exc.TravelException,
			"OutOfTimelineError": exc.OutOfTimelineError,
			"HistoricKeyError": exc.HistoricKeyError,
			"NotInKeyframeError": exc.NotInKeyframeError,
			"WorkerProcessReadOnlyError": exc.WorkerProcessReadOnlyError,
		}

		def unpack_graph(ext: bytes) -> nx.Graph:
			cls, node, adj, graph = unpacker(ext)
			blank = {
				"Graph": nx.Graph,
				"DiGraph": nx.DiGraph,
				"MultiGraph": nx.MultiGraph,
				"MultiDiGraph": nx.MultiDiGraph,
			}[cls]()
			blank._node = node
			blank._adj = adj
			blank.graph = graph
			return blank

		def unpack_exception(ext: bytes) -> Exception:
			data: tuple[str, dict | None] = unpacker(ext)
			if data[0] not in excs:
				return Exception(*data)
			ret = excs[data[0]](*data[2:])
			if data[1] is not None:
				ret.__traceback__ = Traceback.from_dict(data[1]).to_traceback()
			return ret

		def unpack_char(ext: bytes) -> char_cls:
			charn = unpacker(ext)
			return char_cls(self, charn, init_rulebooks=False)

		def unpack_place(ext: bytes) -> place_cls:
			charn, placen = unpacker(ext)
			return place_cls(
				char_cls(self, charn, init_rulebooks=False), placen
			)

		def unpack_thing(ext: bytes) -> thing_cls:
			charn, thingn = unpacker(ext)
			# Breaks if the thing hasn't been instantiated yet, not great
			return self.character[charn].thing[thingn]

		def unpack_portal(ext: bytes) -> portal_cls:
			charn, orign, destn = unpacker(ext)
			return portal_cls(
				char_cls(self, charn, init_rulebooks=False), orign, destn
			)

		def unpack_seq(t: type[_T], ext: bytes) -> _T:
			unpacked = unpacker(ext)
			if not isinstance(unpacked, list):
				raise TypeError("Tried to unpack", type(unpacked), t)
			return t(unpacked)

		def unpack_func(store: FunctionStore, ext: bytes) -> callable:
			unpacked = unpacker(ext)
			if not isinstance(unpacked, str):
				raise TypeError("Tried to unpack as func", type(unpacked))
			return getattr(store, unpacked)

		handlers = {
			MsgpackExtensionType.ellipsis.value: lambda _: ...,
			MsgpackExtensionType.graph.value: unpack_graph,
			MsgpackExtensionType.character.value: unpack_char,
			MsgpackExtensionType.place.value: unpack_place,
			MsgpackExtensionType.thing.value: unpack_thing,
			MsgpackExtensionType.portal.value: unpack_portal,
			MsgpackExtensionType.tuple.value: partial(unpack_seq, tuple),
			MsgpackExtensionType.frozenset.value: partial(
				unpack_seq, frozenset
			),
			MsgpackExtensionType.set.value: partial(unpack_seq, set),
			MsgpackExtensionType.function.value: partial(
				unpack_func, self.function
			),
			MsgpackExtensionType.method.value: partial(
				unpack_func, self.method
			),
			MsgpackExtensionType.trigger.value: partial(
				unpack_func, self.trigger
			),
			MsgpackExtensionType.prereq.value: partial(
				unpack_func, self.prereq
			),
			MsgpackExtensionType.action.value: partial(
				unpack_func, self.action
			),
			MsgpackExtensionType.exception.value: unpack_exception,
		}

		def unpack_handler(
			code: MsgpackExtensionType, data: bytes
		) -> Value | Exception | msgpack.ExtType:
			if code in handlers:
				return handlers[code](data)
			return msgpack.ExtType(code, data)

		def unpacker(b: bytes):
			the_unpacker = msgpack.Unpacker(
				ext_hook=unpack_handler, raw=False, strict_map_key=False
			)
			the_unpacker.feed(b)
			# Deliberately only returning the initial item;
			# others are likely to be null bytes as a result of the
			# way browsers work, and anyway if you really want more
			# you can just pack a list
			return the_unpacker.unpack()

		return unpacker

	@abstractmethod
	def _get_node(
		self, char: AbstractCharacter | CharName, node: NodeName
	) -> Node: ...

	@abstractmethod
	def _btt(self) -> tuple[Branch, Turn, Tick]: ...

	def branches(self) -> KeysView:
		return self._branches_d.keys()

	def branch_parent(self, branch: Branch | None) -> Branch | None:
		if branch is None or branch not in self._branches_d:
			return None
		return self._branches_d[branch][0]

	def _branch_start(self, branch: Branch | None = None) -> tuple[Turn, Tick]:
		if branch is None:
			branch = self.branch
		_, turn, tick, _, _ = self._branches_d[branch]
		return turn, tick

	def _branch_end(self, branch: Branch | None = None) -> tuple[Turn, Tick]:
		if branch is None:
			branch = self.branch
		_, _, _, turn, tick = self._branches_d[branch]
		return turn, tick

	@abstractmethod
	def _start_branch(
		self, parent: Branch, branch: Branch, turn: Turn, tick: Tick
	) -> None: ...

	@abstractmethod
	def _set_btt(self, branch: Branch, turn: Turn, tick: Tick) -> None: ...

	@abstractmethod
	def _extend_branch(
		self, branch: Branch, turn: Turn, tick: Tick
	) -> None: ...

	@abstractmethod
	def load_at(self, branch: Branch, turn: Turn, tick: Tick) -> None: ...

	def branch_start_turn(self, branch: Branch | None = None) -> Turn:
		return self._branch_start(branch)[0]

	def branch_start_tick(self, branch: Branch | None = None) -> Tick:
		return self._branch_start(branch)[1]

	def branch_end_turn(self, branch: Branch | None = None) -> Turn:
		return self._branch_end(branch)[0]

	def branch_end_tick(self, branch: Branch | None = None) -> Tick:
		return self._branch_end(branch)[1]

	@abstractmethod
	def turn_end(
		self, branch: Branch | None = None, turn: Turn | None = None
	) -> Tick: ...

	@abstractmethod
	def turn_end_plan(
		self, branch: Branch | None = None, turn: Turn | None = None
	) -> Tick: ...

	@abstractmethod
	def add_character(
		self,
		name: CharName,
		data: nx.Graph | DiGraph | None = None,
		layout: bool = False,
		node: NodeValDict | None = None,
		edge: EdgeValDict | None = None,
		**kwargs,
	): ...

	def new_character(
		self,
		name: CharName,
		data: nx.Graph | DiGraph | None = None,
		layout: bool = False,
		node: NodeValDict | None = None,
		edge: EdgeValDict | None = None,
		**kwargs,
	):
		self.add_character(name, data)
		return self.character[name]

	def coin_flip(self) -> bool:
		"""Return True or False with equal probability."""
		return self.choice((True, False))

	def die_roll(self, d: Annotated[int, Ge(1)]) -> int:
		"""Roll a die with ``d`` faces. Return the result."""
		return self.randint(1, d)

	def dice(
		self, n: Annotated[int, Ge(1)], d: Annotated[int, Ge(1)]
	) -> Iterable[int]:
		"""Roll ``n`` dice with ``d`` faces, and yield the results.

		This is an iterator. You'll get the result of each die in
		succession.

		"""
		for i in range(0, n):
			yield self.die_roll(d)

	def dice_check(
		self,
		n: Annotated[int, Ge(1)],
		d: Annotated[int, Ge(1)],
		target: Annotated[int, Ge(0)],
		comparator: str | Callable[[int, int], bool] = "<=",
	) -> bool:
		"""Roll ``n`` dice with ``d`` sides, sum them, and compare

		If ``comparator`` is provided, use it instead of the default <=.
		You may use a string like '<' or '>='.

		"""
		from operator import eq, ge, gt, le, lt, ne

		comps: dict[str, Callable] = {
			">": gt,
			"<": lt,
			">=": ge,
			"<=": le,
			"=": eq,
			"==": eq,
			"!=": ne,
		}
		if not callable(comparator):
			comparator = comps[comparator]
		return comparator(sum(self.dice(n, d)), target)

	def chance(self, f: Annotated[float, Ge(0.0), Le(1.0)]) -> bool:
		"""Return True or False with a given unit probability

		Supply a float between 0.0 and 1.0 to express the probability--
		or use `percent_chance`

		"""
		if f <= 0.0:
			return False
		if f >= 1.0:
			return True
		return f > self._rando.random()

	def percent_chance(
		self,
		pct: Annotated[int, Ge(0), Le(100)]
		| Annotated[float, Ge(0.0), Le(100.0)],
	) -> bool:
		"""Return True or False with a given percentile probability

		Values not between 0 and 100 are treated as though they
		were 0 or 100, whichever is nearer.

		"""
		return self.chance(pct / 100)

	betavariate = get_rando("_rando.betavariate")
	choice = get_rando("_rando.choice")
	expovariate = get_rando("_rando.expovariate")
	gammavariate = get_rando("_rando.gammavariate")
	gauss = get_rando("_rando.gauss")
	getrandbits = get_rando("_rando.getrandbits")
	lognormvariate = get_rando("_rando.lognormvariate")
	normalvariate = get_rando("_rando.normalvariate")
	paretovariate = get_rando("_rando.paretovariate")
	randint = get_rando("_rando.randint")
	random = get_rando("_rando.random")
	randrange = get_rando("_rando.randrange")
	sample = get_rando("_rando.sample")
	shuffle = get_rando("_rando.shuffle")
	triangular = get_rando("_rando.triangular")
	uniform = get_rando("_rando.uniform")
	vonmisesvariate = get_rando("_rando.vonmisesvariate")
	weibullvariate = get_rando("_rando.weibullvariate")


class AbstractCharacter(DiGraph):
	"""The Character API, with all requisite mappings and graph generators.

	Mappings resemble those of a NetworkX digraph:

	* ``thing`` and ``place`` are subsets of ``node``
	* ``edge``, ``adj``, and ``succ`` are aliases of ``portal``
	* ``pred`` is an alias to ``preportal``
	* ``stat`` is a dict-like mapping of data that changes over game-time,
	to be used in place of graph attributes

	"""

	engine = getatt("db")
	no_unwrap = True
	name: Hashable
	db: AbstractEngine

	@staticmethod
	def is_directed():
		return True

	@staticmethod
	def is_multigraph():
		return False

	@abstractmethod
	def add_place(self, name: NodeName, **kwargs):
		pass

	def add_node(self, name: NodeName, **kwargs):
		self.add_place(name, **kwargs)

	@abstractmethod
	def add_places_from(self, seq: Iterable, **attrs):
		pass

	def add_nodes_from(self, seq: Iterable, **attrs):
		self.add_places_from(seq, **attrs)

	def new_place(self, name: NodeName, **kwargs):
		"""Add a Place and return it.

		If there's already a Place by that name, put a number on the end.

		"""
		if name not in self.node:
			self.add_place(name, **kwargs)
			return self.place[name]
		if isinstance(name, str):
			n = 0
			while name + str(n) in self.node:
				n += 1
			self.add_place(name + str(n), **kwargs)
			return self.place[name]
		raise KeyError("Already have a node named {}".format(name))

	def new_node(self, name: NodeName, **kwargs):
		return self.new_place(name, **kwargs)

	@abstractmethod
	def add_thing(self, name: NodeName, location: NodeName, **kwargs):
		pass

	@abstractmethod
	def add_things_from(self, seq: Iterable, **attrs):
		pass

	def new_thing(self, name: NodeName, location: NodeName, **kwargs):
		"""Add a Thing and return it.

		If there's already a Thing by that name, put a number on the end.

		"""
		if name not in self.node:
			self.add_thing(name, location, **kwargs)
			return self.thing[name]
		if isinstance(name, str):
			if name in self.node:
				n = 0
				while name + str(n) in self.node:
					n += 1
				name = name + str(n)
			self.add_thing(name, location, **kwargs)
			return self.thing[name]
		raise KeyError("Already have a thing named {}".format(name))

	@abstractmethod
	def place2thing(self, place: NodeName, location: NodeName) -> None: ...

	@abstractmethod
	def thing2place(self, thing: NodeName) -> None: ...

	def remove_node(self, node):
		if node in self.node:
			self.node[node].delete()

	def remove_nodes_from(self, nodes: Iterable[NodeName]):
		for node in nodes:
			if node in self.node:
				self.node[node].delete()

	@abstractmethod
	def add_portal(self, orig: NodeName, dest: NodeName, **kwargs):
		pass

	def add_edge(self, orig: NodeName, dest: NodeName, **kwargs):
		self.add_portal(orig, dest, **kwargs)

	def new_portal(self, orig: NodeName, dest: NodeName, **kwargs):
		self.add_portal(orig, dest, **kwargs)
		return self.portal[orig][dest]

	@abstractmethod
	def add_portals_from(self, seq: Iterable, **attrs):
		pass

	def add_edges_from(self, seq: Iterable, **attrs):
		self.add_portals_from(seq, **attrs)

	@abstractmethod
	def remove_portal(self, origin: NodeName, destination: NodeName):
		pass

	def remove_portals_from(self, seq: Iterable[tuple[NodeName, NodeName]]):
		for orig, dest in seq:
			del self.portal[orig][dest]

	def remove_edges_from(self, seq: Iterable[tuple[NodeName, NodeName]]):
		self.remove_portals_from(seq)

	@abstractmethod
	def remove_place(self, place: NodeName):
		pass

	def remove_places_from(self, seq: Iterable[NodeName]):
		for place in seq:
			self.remove_place(place)

	@abstractmethod
	def remove_thing(self, thing: NodeName) -> None:
		pass

	def remove_things_from(self, seq: Iterable[NodeName]) -> None:
		for thing in seq:
			self.remove_thing(thing)

	@abstractmethod
	def add_unit(
		self, a: CharName | Node, b: Optional[NodeName] = None
	) -> None:
		pass

	@abstractmethod
	def remove_unit(
		self, a: CharName | Node, b: Optional[NodeName] = None
	) -> None:
		pass

	def __eq__(self, other: AbstractCharacter):
		return isinstance(other, AbstractCharacter) and self.name == other.name

	def __iter__(self):
		return iter(self.node)

	def __len__(self):
		return len(self.node)

	def __bool__(self):
		try:
			return self.name in self.db.character
		except AttributeError:
			return False  # we can't "really exist" when we've no engine

	def __contains__(self, k: NodeName):
		return k in self.node

	def __getitem__(self, k: NodeName):
		return self.adj[k]

	ThingMapping: type[SpecialMapping]

	@cached_property
	def thing(self) -> ThingMapping:
		return self.ThingMapping(self)

	PlaceMapping: type[SpecialMapping]

	@cached_property
	def place(self) -> PlaceMapping:
		return self.PlaceMapping(self)

	ThingPlaceMapping: type[SpecialMapping]

	@cached_property
	def _node(self) -> ThingPlaceMapping:
		return self.ThingPlaceMapping(self)

	node: ThingPlaceMapping = getatt("_node")
	nodes: ThingPlaceMapping = getatt("_node")

	PortalSuccessorsMapping: type[SpecialMapping]

	@cached_property
	def _succ(self) -> PortalSuccessorsMapping:
		return self.PortalSuccessorsMapping(self)

	portal: PortalSuccessorsMapping = getatt("_succ")
	adj: PortalSuccessorsMapping = getatt("_succ")
	succ: PortalSuccessorsMapping = getatt("_succ")
	edge: PortalSuccessorsMapping = getatt("_succ")
	_adj: PortalSuccessorsMapping = getatt("_succ")

	PortalPredecessorsMapping: type[SpecialMapping]

	@cached_property
	def _pred(self) -> PortalPredecessorsMapping:
		return self.PortalPredecessorsMapping(self)

	preportal: PortalPredecessorsMapping = getatt("_pred")
	pred: PortalPredecessorsMapping = getatt("_pred")

	UnitGraphMapping: type[SpecialMapping]

	@cached_property
	def unit(self) -> UnitGraphMapping:
		return self.UnitGraphMapping(self)

	stat: GraphMapping = getatt("graph")

	def units(self):
		for units in self.unit.values():
			yield from units.values()

	def historical(self, stat: Stat):
		from .query import EntityStatAlias

		return EntityStatAlias(entity=self.stat, stat=stat)

	def do(self, func: callable, *args, **kwargs) -> AbstractCharacter:
		"""Apply the function to myself, and return myself.

		Look up the function in the method store if needed. Pass it any
		arguments given, keyword or positional.

		Useful chiefly when chaining.

		"""
		if not callable(func):
			func = getattr(self.engine.method, func)
		func(self, *args, **kwargs)
		return self

	def copy_from(self, g: AbstractCharacter) -> None:
		"""Copy all nodes and edges from the given graph into this.

		Return myself.

		"""
		renamed = {}
		for k in g.nodes:
			ok = k
			if k in self.place:
				n = 0
				while k in self.place:
					k = ok + (n,) if isinstance(ok, tuple) else (ok, n)
					n += 1
			renamed[ok] = k
			self.place[k] = g.nodes[k]
		if type(g) is nx.MultiDiGraph:
			g = nx.DiGraph(g)
		elif type(g) is nx.MultiGraph:
			g = nx.Graph(g)
		if type(g) is nx.DiGraph:
			for u, v in g.edges:
				self.edge[renamed[u]][renamed[v]] = g.adj[u][v]
		else:
			assert type(g) is nx.Graph
			for u, v, d in g.edges.data():
				self.add_portal(renamed[u], renamed[v], symmetrical=True, **d)
		return self

	def become(self, g: AbstractCharacter) -> AbstractCharacter:
		"""Erase all my nodes and edges. Replace them with a copy of the graph
		provided.

		Return myself.

		"""
		self.clear()
		self.place.update(g.nodes)
		self.adj.update(g.adj)
		return self

	def clear(self) -> None:
		self.node.clear()
		self.portal.clear()
		self.stat.clear()

	def _lookup_comparator(self, comparator: callable | str) -> callable:
		if callable(comparator):
			return comparator
		ops = {"ge": ge, "gt": gt, "le": le, "lt": lt, "eq": eq}
		if comparator in ops:
			return ops[comparator]
		return getattr(self.engine.function, comparator)

	def cull_nodes(
		self,
		stat: Stat,
		threshold: float = 0.5,
		comparator: callable | str = ge,
	) -> None:
		"""Delete nodes whose stat >= ``threshold`` (default 0.5).

		Optional argument ``comparator`` will replace >= as the test
		for whether to cull. You can use the name of a stored function.

		"""
		comparator = self._lookup_comparator(comparator)
		dead = [
			name
			for name, node in self.node.items()
			if stat in node and comparator(node[stat], threshold)
		]
		self.remove_nodes_from(dead)
		return self

	def cull_portals(
		self,
		stat: Stat,
		threshold: float = 0.5,
		comparator: callable | str = ge,
	):
		"""Delete portals whose stat >= ``threshold`` (default 0.5).

		Optional argument ``comparator`` will replace >= as the test
		for whether to cull. You can use the name of a stored function.

		"""
		comparator = self._lookup_comparator(comparator)
		dead = []
		for u in self.portal:
			for v in self.portal[u]:
				if stat in self.portal[u][v] and comparator(
					self.portal[u][v][stat], threshold
				):
					dead.append((u, v))
		self.remove_edges_from(dead)
		return self

	cull_edges = cull_portals


DiGraph.register(AbstractCharacter)


def normalize_layout(l: dict) -> dict:
	"""Make sure all the spots in a layout are where you can click.

	Returns a copy of the layout with all spot coordinates are
	normalized to within (0.0, 0.98).

	"""
	xs = []
	ys = []
	ks = []
	for k, (x, y) in l.items():
		xs.append(x)
		ys.append(y)
		ks.append(k)
	minx = np.min(xs)
	maxx = np.max(xs)
	if maxx == minx:
		xnorm = np.array([0.5] * len(xs))
	else:
		xco = 0.98 / (maxx - minx)
		xnorm = np.multiply(np.subtract(xs, [minx] * len(xs)), xco)
	miny = np.min(ys)
	maxy = np.max(ys)
	if miny == maxy:
		ynorm = np.array([0.5] * len(ys))
	else:
		yco = 0.98 / (maxy - miny)
		ynorm = np.multiply(np.subtract(ys, [miny] * len(ys)), yco)
	return dict(zip(ks, zip(map(float, xnorm), map(float, ynorm))))


class AbstractThing(MutableMapping):
	character: AbstractCharacter
	engine: AbstractEngine
	name: NodeName

	@property
	def location(self) -> Node:
		"""The ``Thing`` or ``Place`` I'm in."""
		locn = self["location"]
		if locn is None:
			raise AttributeError("Not really a Thing")
		try:
			return self.engine._get_node(self.character, locn)
		except KeyError as ex:
			raise AttributeError("Doesn't really exist") from ex

	@location.setter
	def location(self, v: Node | NodeName):
		if hasattr(v, "name"):
			v = v.name
		self["location"] = v

	def go_to_place(
		self, place: Node | NodeName, weight: Optional[Stat] = None
	) -> int:
		"""Assuming I'm in a node that has a :class:`Portal` direct
		to the given node, schedule myself to travel to the
		given :class:`Place`, taking an amount of time indicated by
		the ``weight`` stat on the :class:`Portal`, if given; else 1
		turn.

		Return the number of turns the travel will take.

		"""
		if hasattr(place, "name"):
			placen = place.name
		else:
			placen = place
		curloc = self["location"]
		orm = self.character.engine
		turns = (
			1
			if weight is None
			else self.engine._portal_objs[
				(self.character.name, curloc, place)
			].get(weight, 1)
		)
		with self.engine.plan():
			orm.turn += turns
			self["location"] = placen
		return turns

	def follow_path(
		self,
		path: list[NodeName],
		weight: Optional[Stat] = None,
		check: bool = True,
	) -> int:
		"""Go to several nodes in succession, deciding how long to
		spend in each by consulting the ``weight`` stat of the
		:class:`Portal` connecting the one node to the next,
		default 1 turn.

		Return the total number of turns the travel will take. Raise
		:class:`TravelException` if I can't follow the whole path,
		either because some of its nodes don't exist, or because I'm
		scheduled to be somewhere else. Set ``check=False`` if
		you're really sure the path is correct, and this function
		will be faster.

		"""
		if len(path) < 2:
			raise ValueError("Paths need at least 2 nodes")
		eng = self.character.engine
		if check:
			prevplace = path.pop(0)
			if prevplace != self["location"]:
				raise ValueError("Path does not start at my present location")
			subpath = [prevplace]
			for place in path:
				if (
					prevplace not in self.character.portal
					or place not in self.character.portal[prevplace]
				):
					raise exc.TravelException(
						"Couldn't follow portal from {} to {}".format(
							prevplace, place
						),
						path=subpath,
						traveller=self,
					)
				subpath.append(place)
				prevplace = place
		else:
			subpath = path.copy()
		turns_total = 0
		prevsubplace = subpath.pop(0)
		turn_incs = []
		branch, turn, tick = eng._btt()
		for subplace in subpath:
			if weight is not None:
				turn_incs.append(
					self.engine._edge_val_cache.retrieve(
						self.character.name,
						prevsubplace,
						subplace,
						0,
						branch,
						turn,
						tick,
					)
				)
			else:
				turn_incs.append(1)
			turns_total += turn_incs[-1]
			turn += turn_incs[-1]
			tick = eng._turn_end_plan.get((branch, turn), 0)
		with eng.plan():
			for subplace, turn_inc in zip(subpath, turn_incs):
				eng.turn += turn_inc
				self["location"] = subplace
		return turns_total

	def travel_to(
		self,
		dest: Node | NodeName,
		weight: Optional[Stat] = None,
		graph: nx.DiGraph = None,
	) -> int:
		"""Find the shortest path to the given node from where I am
		now, and follow it.

		If supplied, the ``weight`` stat of each :class:`Portal` along
		the path will be used in pathfinding, and for deciding how
		long to stay in each Place along the way. Otherwise, I will stay
		in each :class:`Place` for 1 turn.

		The ``graph`` argument may be any NetworkX-style graph. It
		will be used for pathfinding if supplied, otherwise I'll use
		my :class:`Character`. In either case, however, I will attempt
		to actually follow the path using my :class:`Character`, which
		might not be possible if the supplied ``graph`` and my
		:class:`Character` are too different. If it's not possible,
		I'll raise a :class:`TravelException`, whose ``subpath``
		attribute holds the part of the path that I *can* follow. To
		make me follow it, pass it to my ``follow_path`` method.

		Return value is the number of turns the travel will take.

		"""
		destn = dest.name if hasattr(dest, "name") else dest
		if destn == self.location.name:
			raise ValueError("I'm already there", self.name, destn)
		graph = self.character if graph is None else graph
		path = nx.shortest_path(graph, self["location"], destn, weight)
		return self.follow_path(path, weight)


def repr_call_sig(func: callable | str, *args, **kwargs):
	if not isinstance(func, (str, bytes)):
		func = func.__name__
	return (
		f"{func}({', '.join(map(repr, args))}"
		f"{', ' if args and kwargs else ''}"
		f"{', '.join(f'{arg}={repr(item)}' for (arg, item) in kwargs.items())})"
	)


def print_call_sig(
	func: callable | str, *args, file=sys.stdout, end="\n", **kwargs
):
	print(repr_call_sig(func, *args, **kwargs), file=file, end=end)


@contextmanager
def _garbage_ctx(collect=True):
	"""Context manager to disable the garbage collector

	:param collect: Whether to immediately collect garbage upon context exit

	"""
	gc_was_active = gc.isenabled()
	if gc_was_active:
		gc.disable()
	yield
	if gc_was_active:
		gc.enable()
	if collect:
		gc.collect()


def _garbage_dec(fn: callable, collect=True) -> callable:
	"""Decorator to disable the garbage collector for a function

	:param collect: Whether to immediately collect garbage when the function returns

	"""

	@wraps(fn)
	def garbage(*args, **kwargs):
		with _garbage_ctx(collect=collect):
			return fn(*args, **kwargs)

	return garbage


def garbage(arg: callable = None, collect: bool = False):
	"""Disable the garbage collector, then re-enable it when done.

	May be used as a context manager or a decorator.

	:param collect: Whether to immediately run a collection after re-enabling
		the garbage collector. Default ``False``.

	"""

	if arg is None:
		return _garbage_ctx(collect=collect)
	else:
		return _garbage_dec(arg, collect=collect)


def world_locked(fn: callable) -> callable:
	"""Decorator for functions that alter the world state

	They will hold a reentrant lock, preventing more than one function
	from mutating the world at a time.

	"""

	@wraps(fn)
	def lockedy(*args, **kwargs):
		with args[0].world_lock:
			return fn(*args, **kwargs)

	return lockedy


class TimeSignal(Signal, Sequence):
	"""Acts like a tuple of ``(branch, turn, tick)`` for the most part.

	This is a ``Signal``. To set a function to be called whenever the
	branch or turn changes, pass it to my ``connect`` method.

	"""

	engine: AbstractEngine

	def __init__(self, engine: AbstractEngine):
		super().__init__()
		self.engine = engine

	def __iter__(self):
		yield self.engine.branch
		yield self.engine.turn
		yield self.engine.tick

	def __len__(self):
		return 3

	def __getitem__(self, i: str | int) -> Branch | Turn | Tick:
		if i in ("branch", 0):
			return self.engine.branch
		if i in ("turn", 1):
			return self.engine.turn
		if i in ("tick", 2):
			return self.engine.tick
		if isinstance(i, int):
			raise IndexError(i)
		else:
			raise KeyError(i)

	def __setitem__(self, i: str | int, v: str | int) -> None:
		if i in ("branch", 0):
			self.engine.branch = Branch(v)
		elif i in ("turn", 1):
			self.engine.turn = Turn(v)
		elif i in ("tick", 2):
			self.engine.tick = Tick(v)
		else:
			exctyp = KeyError if isinstance(i, str) else IndexError
			raise exctyp(i)

	def __str__(self):
		return str(tuple(self))

	def __eq__(self, other):
		return tuple(self) == other

	def __ne__(self, other):
		return tuple(self) != other

	def __gt__(self, other):
		return tuple(self) > other

	def __ge__(self, other):
		return tuple(self) >= other

	def __lt__(self, other):
		return tuple(self) < other

	def __le__(self, other):
		return tuple(self) <= other


class TimeSignalDescriptor:
	__doc__ = TimeSignal.__doc__

	def __get__(self, inst, cls):
		if not hasattr(inst, "_time_signal"):
			inst._time_signal = TimeSignal(inst)
		return inst._time_signal

	def __set__(self, inst: AbstractEngine, val: Time):
		if getattr(inst, "_worker", False):
			raise WorkerProcessReadOnlyError(
				"Tried to change the world state in a worker process"
			)
		if not hasattr(inst, "_time_signal"):
			inst._time_signal = TimeSignal(inst)
		sig = inst._time_signal
		branch_then, turn_then, tick_then = inst._btt()
		branch_now, turn_now, tick_now = val
		if (branch_then, turn_then, tick_then) == (
			branch_now,
			turn_now,
			tick_now,
		):
			return
		e = inst
		# enforce the arrow of time, if it's in effect
		if (
			hasattr(e, "_forward")
			and e._forward
			and hasattr(e, "_planning")
			and not e._planning
		):
			if branch_now != branch_then:
				raise TimeError("Can't change branches in a forward context")
			if turn_now < turn_then:
				raise TimeError(
					"Can't time travel backward in a forward context"
				)
			if turn_now > turn_then + 1:
				raise TimeError("Can't skip turns in a forward context")
		# make sure I'll end up within the revision range of the
		# destination branch

		if branch_now in e.branches():
			e._extend_branch(branch_now, turn_now, tick_now)
			e.load_at(branch_now, turn_now, tick_now)
		else:
			e._start_branch(branch_then, branch_now, turn_now, tick_now)
		e._set_btt(branch_now, turn_now, tick_now)
		sig.send(
			e,
			branch_then=branch_then,
			turn_then=turn_then,
			tick_then=tick_then,
			branch_now=branch_now,
			turn_now=turn_now,
			tick_now=tick_now,
		)



import os
import shutil
import tempfile

from lisien.engine import Engine
from lisien.examples.college import install

outpath = os.path.join(
	os.path.abspath(os.path.dirname(__file__)), "college24_premade.tar.xz"
)
if os.path.exists(outpath):
	os.remove(outpath)
with tempfile.TemporaryDirectory() as directory:
	with Engine(
		directory,
		workers=0,
		keep_rules_journal=False,
		commit_interval=1,
		connect_string=f"sqlite:///{directory}/world.sqlite3",
	) as eng:
		install(eng)
		for i in range(24):
			print(i)
			eng.next_turn()
		print("Done simulating.")
	print("Compressing...")
	shutil.make_archive(outpath[:-7], "xztar", directory, ".")
print("All done")



# This file is part of Elide, frontend to Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial
from unittest.mock import call, patch

import networkx as nx
import pytest

from lisien.engine import Engine
from lisien.examples.kobold import inittest
from lisien.tests.util import make_test_engine

testgraphs = [nx.chvatal_graph()]
# have to name it after creation because it clears the create_using
path_graph_9 = nx.path_graph(9)
path_graph_9.name = "path_graph_9"
testgraphs.append(path_graph_9)


@pytest.fixture
def db(tmp_path, execution, database):
	with make_test_engine(tmp_path, execution, database) as orm:
		for graph in testgraphs:
			orm.new_character(graph.name, graph)
			if not graph.is_directed():
				graph = nx.to_directed(graph)
			assert set(graph.nodes.keys()) == set(
				orm.character[graph.name].nodes.keys()
			), "{}'s nodes changed during instantiation".format(graph.name)
			assert set(graph.edges) == set(
				orm.character[graph.name].edges.keys()
			), "{}'s edges changed during instantiation".format(graph.name)
		yield orm


def test_basic_load(db):
	for graph in testgraphs:
		if not graph.is_directed():
			graph = nx.to_directed(graph)
		alleged = db.character[graph.name]
		assert set(graph.nodes.keys()) == set(alleged.nodes.keys()), (
			"{}'s nodes are not the same after load".format(graph.name)
		)
		assert set(graph.edges) == set(alleged.edges), (
			"{}'s edges are not the same after load".format(graph.name)
		)


def test_keyframe_load(db):
	for graph in testgraphs:
		nodes_kf = db._nodes_cache.keyframe
		assert (graph.name,) in nodes_kf, "{} not in nodes cache".format(
			graph.name
		)
		assert "trunk" in nodes_kf[graph.name,], (
			"trunk branch not in nodes cache for {}".format(graph.name)
		)
		assert nodes_kf[graph.name,]["trunk"].rev_gettable(0), (
			"turn 0 not in nodes cache for {}".format(graph.name)
		)
		assert nodes_kf[graph.name,]["trunk"][0].rev_gettable(0), (
			"tick 0 not in nodes cache for {}".format(graph.name)
		)
		assert db._nodes_cache.keyframe[graph.name,]["trunk"][0][0] == {
			node: True for node in graph.nodes.keys()
		}, "{} not loaded correctly, got {}".format(
			graph.name, nodes_kf["trunk"][0][0]
		)
		edges_kf = db._edges_cache.keyframe
		for orig in graph.adj:
			for dest in graph.adj[orig]:
				assert (
					graph.name,
					orig,
					dest,
				) in edges_kf, "{} not in edges cache".format(
					(graph.name, orig, dest)
				)
				this_edge = edges_kf[graph.name, orig, dest]
				assert "trunk" in this_edge, (
					"trunk branch not in edges cache for {}".format(
						(graph.name, orig, dest)
					)
				)
				assert this_edge["trunk"].rev_gettable(0), (
					"turn 0 not in trunk branch of edges cache for {}".format(
						(graph.name, orig, dest)
					)
				)
				assert this_edge["trunk"][0].rev_gettable(0), (
					"tick 0 not in turn 0 of trunk branch of edges cache for {}".format(
						(graph.name, orig, dest)
					)
				)
				assert db._edges_cache.keyframe[graph.name, orig, dest][
					"trunk"
				][0][0] == {0: True}, "{} not loaded".format(
					(graph.name, orig, dest)
				)
		for node, vals in graph.nodes.items():
			assert (
				db._node_val_cache.keyframe[graph.name, node]["trunk"][0][0]
				== vals
			)
		for edge in graph.edges:
			assert (
				db._edge_val_cache.keyframe[(graph.name, *edge)]["trunk"][0][0]
				== graph.edges[edge]
			)


def test_keyframe_unload(tmp_path, execution, non_null_database):
	# TODO: test edge cases involving tick-precise unloads
	eng = partial(make_test_engine, tmp_path, execution, non_null_database)
	with eng() as orm:
		g = orm.new_character("g", nx.grid_2d_graph(3, 3))
		orm.next_turn()
		assert orm.turn == 1
		assert (
			"g",
			(0, 0),
			(0, 1),
		) in orm._edges_cache.keyframe and 0 in orm._edges_cache.keyframe[
			"g", (0, 0), (0, 1)
		]["trunk"]
		del g.node[1, 1]
		g.add_node("a")
		g.add_edge((0, 0), "a")
		orm.next_turn()
		assert orm.turn == 2
		orm.snap_keyframe()
		g.add_node((4, 4))
		g.add_edge((3, 3), (4, 4))
		assert (
			"g",
			(0, 0),
			(0, 1),
		) in orm._edges_cache.keyframe and 0 in orm._edges_cache.keyframe[
			"g", (0, 0), (0, 1)
		]["trunk"]
		assert (
			("g",) in orm._nodes_cache.keyframe
			and "trunk" in orm._nodes_cache.keyframe["g",]
			and 0 in orm._nodes_cache.keyframe["g",]["trunk"]
		)
		orm.unload()
		assert not orm._time_is_loaded("trunk", 1)
		if "trunk" in orm._nodes_cache.keyframe["g",]:
			assert 0 not in orm._nodes_cache.keyframe["g",]["trunk"]
		assert ("g", (0, 0), (0, 1)) in orm._edges_cache.keyframe
		assert "trunk" in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]
		assert 2 in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]["trunk"]
		assert 0 not in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]["trunk"]
		endtick = orm.tick
	with eng() as orm:
		assert not orm._time_is_loaded("trunk", 1)
		assert orm._time_is_loaded("trunk", 2, endtick)
		assert ("g", (0, 0), (0, 1)) in orm._edges_cache.keyframe
		assert 2 in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]["trunk"]
		assert 0 not in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]["trunk"]
		g = orm.character["g"]
		if "trunk" in orm._nodes_cache.keyframe["g",]:
			assert 0 not in orm._nodes_cache.keyframe["g",]["trunk"]
		if (
			("g", (0, 0), (0, 1)) in orm._edges_cache.keyframe
			and "trunk" in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]
		):
			assert (
				0
				not in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]["trunk"]
			)
		assert not orm._time_is_loaded("trunk", 1)
		orm.turn = 0
		assert orm._time_is_loaded("trunk", 1)
		assert 0 in orm._edges_cache.keyframe["g", (0, 0), (0, 1)]["trunk"]
		orm.branch = "u"
		del g.node[1, 2]
		orm.unload()
	with eng() as orm:
		assert orm.branch == "u"
		assert (
			("g", (1, 1), (1, 2)) not in orm._edges_cache.keyframe
			or "trunk" not in orm._edges_cache.keyframe["g", (1, 1), (1, 2)]
		)
		g = orm.character["g"]
		assert (1, 2) not in g.nodes
		orm.branch = "trunk"
		assert (1, 2) in g.nodes
		assert (
			("g", (1, 1), (1, 2)) in orm._edges_cache.keyframe
			and "trunk" in orm._edges_cache.keyframe["g", (1, 1), (1, 2)]
		)


def test_keyframe_load_init(tmp_path):
	"""Can load a keyframe at start of branch, including locations"""
	eng = Engine(tmp_path, workers=0)
	inittest(eng)
	eng.branch = "new"
	# eng.snap_keyframe()
	eng.close()
	eng = Engine(tmp_path, workers=0)
	# the graphs keyframe is coming up empty
	assert "kobold" in eng.character["physical"].thing
	assert (0, 0) in eng.character["physical"].place
	assert (0, 1) in eng.character["physical"].portal[0, 0]
	eng.close()


def test_multi_keyframe(tmp_path):
	eng = Engine(
		tmp_path, enforce_end_of_time=False, keyframe_on_close=False, workers=0
	)
	inittest(eng)
	eng.snap_keyframe()
	tick0 = eng.tick
	eng.turn = 1
	del eng.character["physical"].place[3, 3]
	eng.snap_keyframe()
	tick1 = eng.tick
	assert ("physical",) in eng._nodes_cache.keyframe
	assert "trunk" in eng._nodes_cache.keyframe["physical",]
	assert 1 in eng._nodes_cache.keyframe["physical",]["trunk"]
	assert tick1 in eng._nodes_cache.keyframe["physical",]["trunk"][1]
	assert (1, 1) in eng._nodes_cache.keyframe["physical",]["trunk"][1][tick1]
	assert (3, 3) not in eng._nodes_cache.keyframe["physical",]["trunk"][1][
		tick1
	]
	eng.close()
	eng = Engine(tmp_path, keyframe_on_close=False, workers=0)
	eng.load_at("trunk", 0, tick0)
	assert eng._time_is_loaded("trunk", 0, tick0)
	assert eng._time_is_loaded("trunk", 0, tick0 + 1)
	assert eng._time_is_loaded("trunk", 1, tick1 - 1)
	assert eng._time_is_loaded("trunk", 1, tick1)
	eng.close()


def test_keyframe_load_unload(tmp_path, non_null_database):
	"""All caches can load and unload before and after kfs"""
	if non_null_database == "sqlite":
		connect_str = f"sqlite:///{tmp_path}/world.sqlite3"
	else:
		connect_str = None
	with Engine(
		tmp_path,
		enforce_end_of_time=False,
		keyframe_on_close=False,
		workers=0,
		connect_string=connect_str,
	) as eng:
		eng.snap_keyframe()
		eng.turn = 1
		inittest(eng)
		eng.snap_keyframe()
		eng.turn = 2
		eng.universal["hi"] = "hello"
		now = eng._btt()
	with Engine(
		tmp_path,
		enforce_end_of_time=False,
		keyframe_on_close=False,
		workers=0,
		connect_string=connect_str,
	) as eng:
		assert eng._time_is_loaded(*now)
		assert not eng._time_is_loaded("trunk", 0)
		eng.turn = 1
		eng.tick = 0
		assert eng._time_is_loaded("trunk", 1)
		assert eng._time_is_loaded("trunk", 1, 0)
		assert eng._time_is_loaded(*now)
		eng.unload()
		assert eng._time_is_loaded("trunk", 1, 0)
		assert not eng._time_is_loaded(*now)
		eng.turn = 2
		eng.branch = "haha"
		eng.snap_keyframe()
		eng.unload()
		assert not eng._time_is_loaded("trunk")


@pytest.fixture
def some_state(tmp_path, non_null_database):
	with Engine(
		tmp_path,
		workers=0,
		random_seed=0,
		connect_string=f"sqlite:///{tmp_path}/world.sqlite3"
		if non_null_database == "sqlite"
		else None,
	) as eng:
		initial_state = nx.DiGraph(
			{
				0: {1: {"omg": "lol"}},
				1: {0: {"omg": "blasphemy"}},
				2: {},
				3: {},
				"it": {},
			}
		)
		initial_state.nodes()[2]["hi"] = "hello"
		initial_state.nodes()["it"]["location"] = 0
		initial_state.graph["wat"] = "nope"
		phys = eng.new_character("physical", initial_state)
		eng.add_character("pointless")
		kf0 = eng.snap_keyframe()
		del kf0["universal"]["rando_state"]
		eng.branch = "b"
		kf1 = eng.snap_keyframe()
		del kf1["universal"]["rando_state"]
		assert kf0 == kf1
		del phys.portal[1][0]
		port = phys.new_portal(0, 2)
		port["hi"] = "bye"
		phys.place[1]["wtf"] = "bbq"
		phys.thing["it"].location = phys.place[1]
		del phys.place[3]
		eng.add_character("pointed")
		del eng.character["pointless"]
		assert "pointless" not in eng.character, "Failed to delete character"
		phys.portal[0][1]["meaning"] = 42
		del phys.portal[0][1]["omg"]
		eng.branch = "trunk"
	return tmp_path


def test_load_branch_to_end(some_state, non_null_database):
	with Engine(
		some_state,
		workers=0,
		random_seed=0,
		connect_string=f"sqlite:///{some_state}/world.sqlite3"
		if non_null_database == "sqlite"
		else None,
	) as eng:
		assert eng.turn == 0
		phys = eng.character["physical"]
		assert 3 in phys.place
		assert phys.portal[1][0]["omg"] == "blasphemy"
		eng.branch = "b"
		assert 3 not in phys.place
		assert 0 not in phys.portal[1]
		assert 2 in phys.portal[0]
		assert phys.portal[0][2]["hi"] == "bye"
		assert phys.place[1]["wtf"] == "bbq"
		assert phys.thing["it"].location == phys.place[1]
		assert "pointless" not in eng.character, "Loaded deleted character"
		assert "pointed" in eng.character
		assert phys.portal[0][1]["meaning"] == 42
		assert "omg" not in phys.portal[0][1]



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import networkx as nx
import pytest


@pytest.fixture(scope="function")
def something(engine):
	yield (
		engine.new_character("physical")
		.new_place("somewhere")
		.new_thing("something")
	)


def test_contents(something):
	pl1 = something.character.place["somewhere"]
	pl2 = something.character.new_place("somewhere2")
	assert something.location == something.character.node["somewhere"]
	assert something.name in pl1.content
	assert something.name not in pl2.content
	assert [something] == list(pl1.contents())
	assert [] == list(pl2.contents())


def test_future_contents(something):
	engine = something.engine
	somewhere = something.location
	with engine.plan():
		engine.turn = 1
		something.location = None
		engine.turn = 2
		somebody = somewhere.new_thing("somebody")
		engine.turn = 0
		someone = somewhere.new_thing("someone")
		sometick = engine.tick
	assert engine.turn == 0
	assert len(somewhere.contents()) == 1
	assert something in somewhere.contents()
	assert someone not in somewhere.contents()
	engine.tick = sometick
	assert len(somewhere.contents()) == 2
	assert something in somewhere.contents()
	assert somebody not in somewhere.contents()
	assert someone in somewhere.contents()
	engine.turn = 1
	engine.tick = engine.turn_end_plan()
	assert len(somewhere.contents()) == 1
	assert something not in somewhere.contents()
	assert someone in somewhere.contents()
	assert somebody not in somewhere.contents()
	engine.turn = 2
	engine.tick = engine.turn_end_plan()
	assert len(somewhere.contents()) == 2
	assert somebody in somewhere.contents()
	assert someone in somewhere.contents()
	assert something not in somewhere.contents()


def test_travel(engine):
	phys = engine.new_character("physical", data=nx.grid_2d_graph(8, 8))
	del phys.place[1, 1]
	del phys.place[6, 1]
	thing1 = phys.place[0, 0].new_thing(1)
	thing2 = phys.place[7, 0].new_thing(2)
	thing1.travel_to(phys.place[7, 7])
	thing2.travel_to(phys.place[0, 7])
	for _ in range(15):
		engine.next_turn()
	assert thing1.location == phys.place[7, 7]
	assert thing2.location == phys.place[0, 7]
	thing1.go_to_place(phys.place[6, 7])
	thing2.go_to_place(phys.place[1, 7])
	engine.next_turn()
	engine.next_turn()
	assert thing1.location == phys.place[6, 7]
	assert thing2.location == phys.place[1, 7]


def test_neighbors_no_successors(sqleng):
	phys = sqleng.new_character("physical")
	here = phys.new_place("here")
	phys.new_place("there")
	assert list(here.neighbors()) == []



from lisien import Engine


def make_test_engine_kwargs(
	path,
	execution,
	database,
	random_seed=69105,
	enforce_end_of_time=False,
):
	kwargs = {
		"random_seed": random_seed,
		"enforce_end_of_time": enforce_end_of_time,
		"prefix": None if database == "null" else path,
	}
	if database == "sqlite":
		kwargs["connect_string"] = f"sqlite:///{path}/world.sqlite3"
	kwargs["workers"] = 2 if execution == "parallel" else 0
	return kwargs


def make_test_engine(path, execution, database):
	return Engine(**make_test_engine_kwargs(path, execution, database))



import pytest

from lisien import Engine
from lisien.proxy.handle import EngineHandle


@pytest.fixture
def handle_empty(tmp_path, database):
	Engine(
		tmp_path if database == "parquetdb" else None,
		workers=0,
		connect_string=f"sqlite:///{tmp_path}/world.sqlite3"
		if database == "sqlite"
		else None,
	).close()
	handle = EngineHandle(
		tmp_path if database == "parquetdb" else None,
		workers=0,
		connect_string=f"sqlite:///{tmp_path}/world.sqlite3"
		if database == "sqlite"
		else None,
	)
	yield handle
	handle.close()


def test_language(handle_empty):
	assert handle_empty.get_language() == "eng"
	handle_empty.set_string("foo", "bar")
	assert handle_empty.get_string_lang_items("eng") == [("foo", "bar")]
	handle_empty.set_language("esp")
	assert handle_empty.get_language() == "esp"
	assert handle_empty.get_string_lang_items("esp") == []
	assert handle_empty.get_string_lang_items() == []
	handle_empty.set_language("eng")
	assert handle_empty.get_string_lang_items() == [("foo", "bar")]
	assert handle_empty.strings_copy() == {"foo": "bar"}
	handle_empty.del_string("foo")
	handle_empty.set_language("esp")
	assert handle_empty.strings_copy("eng") == {}


def test_eternal(handle_empty, database):
	unpack = handle_empty.unpack
	assert unpack(handle_empty.get_eternal("_lisien_schema_version")) == 1
	assert unpack(handle_empty.get_eternal("main_branch")) == "trunk"
	assert unpack(handle_empty.get_eternal("language")) == "eng"
	handle_empty.set_eternal("haha", "lol")
	assert unpack(handle_empty.get_eternal("haha")) == "lol"
	handle_empty.del_eternal("branch")
	with pytest.raises(KeyError):
		handle_empty.get_eternal("branch")
	assert handle_empty.eternal_copy() == {
		b"\xb6_lisien_schema_version": b"\x01",
		b"\xabmain_branch": b"\xa5trunk",
		b"\xa4turn": b"\x00",
		b"\xa4tick": b"\x00",
		b"\xa8language": b"\xa3eng",
		b"\xa4haha": b"\xa3lol",
	}


def test_universal(handle_empty):
	handle_empty.set_universal("foo", "bar")
	handle_empty.set_universal("spam", "tasty")
	univ = handle_empty.snap_keyframe()["universal"]
	assert univ["foo"] == "bar"
	assert univ["spam"] == "tasty"
	handle_empty.del_universal("foo")
	univ = handle_empty.snap_keyframe()["universal"]
	assert "foo" not in univ
	assert univ["spam"] == "tasty"


def test_character(handle_initialized):
	origtime = handle_initialized.get_btt()
	handle_initialized.next_turn()
	handle_initialized.add_character(
		"hello",
		node={
			"hi": {"yes": "very yes"},
			"hello": {"you": "smart"},
			"morning": {"good": 100},
			"salutations": {},
			"me": {"location": "hi"},
		},
		edge={"hi": {"hello": {"good": "morning"}}},
		stat="also",
	)
	assert handle_initialized.node_exists("hello", "hi")
	handle_initialized.set_character_stat("hello", "stoat", "bitter")
	handle_initialized.del_character_stat("hello", "stat")
	handle_initialized.set_node_stat("hello", "hi", "no", "very no")
	handle_initialized.del_node_stat("hello", "hi", "yes")
	handle_initialized.del_character("physical")
	handle_initialized.del_node("hello", "salutations")
	handle_initialized.update_nodes(
		"hello",
		{"hi": {"tainted": True}, "bye": {"toodles": False}, "morning": None},
	)
	handle_initialized.set_thing(
		"hello", "evening", {"location": "bye", "moon": 1.0}
	)
	handle_initialized.add_thing(
		"hello", "moon", "evening", {"phase": "waxing gibbous"}
	)
	handle_initialized.character_set_node_predecessors(
		"hello", "bye", {"hi": {"is-an-edge": True}}
	)
	handle_initialized.add_thing("hello", "neal", "hi", {})
	handle_initialized.add_character("astronauts", {}, {})
	handle_initialized.add_unit("astronauts", "hello", "neal")
	handle_initialized.set_character_rulebook("astronauts", "nasa")
	handle_initialized.set_thing_location("hello", "neal", "moon")
	handle_initialized.set_place("hello", "earth", {})
	handle_initialized.add_portal("hello", "moon", "earth", {})
	assert handle_initialized.thing_travel_to("hello", "neal", "earth") == 1
	kf0 = handle_initialized.snap_keyframe()
	del kf0["universal"]
	assert kf0 == handle_initialized.keyframe[0]
	desttime = handle_initialized.get_btt()
	handle_initialized.time_travel(*origtime)
	kf1 = handle_initialized.snap_keyframe()
	del kf1["universal"]
	assert kf1 == handle_initialized.keyframe[1]
	handle_initialized.time_travel(*desttime)
	kf2 = handle_initialized.snap_keyframe()
	del kf2["universal"]
	assert kf2 == kf0



import pytest


@pytest.fixture(params=["branch-delta", "slow-delta"])
def codepath(request):
	return request.param


def test_character_existence_delta(null_engine, codepath):
	eng = null_engine
	eng.add_character(1)
	eng.add_character(2)
	eng.next_turn()
	eng.add_character(3)
	if codepath == "slow-delta":
		eng.branch = "branch"
	else:
		eng.next_turn()
	del eng.character[2]
	eng.add_character(4)
	delta0 = eng.get_delta(("trunk", 0, 0), tuple(eng.time))
	assert 3 in delta0 and delta0[3] == {
		"character_place_rulebook": ("character_place_rulebook", 3),
		"character_portal_rulebook": ("character_portal_rulebook", 3),
		"character_rulebook": ("character_rulebook", 3),
		"character_thing_rulebook": ("character_thing_rulebook", 3),
		"unit_rulebook": ("unit_rulebook", 3),
	}
	assert 2 in delta0 and delta0[2] is ...
	delta1 = eng.get_delta(
		("trunk", 1, 1),
		("branch", 1) if codepath == "slow-delta" else ("trunk", eng.turn),
	)
	assert 2 in delta1
	assert delta1[2] is ...


def test_unit_delta(null_engine, codepath):
	eng = null_engine
	one = eng.new_character(1)
	six = one.new_place(6)
	seven = six.new_thing(7)
	one.new_place(8)
	two = eng.new_character(2)
	three = two.new_place(3)
	four = three.new_thing(4)
	two.add_place(5)
	one.add_unit(three)
	one.add_unit(four)
	time_a = tuple(eng.time)
	if codepath == "slow-delta":
		eng.branch = "branch"
	else:
		eng.next_turn()
	two.add_unit(six)
	two.add_unit(seven)
	one.remove_unit(three)
	one.remove_unit(four)
	time_b = tuple(eng.time)
	delta0 = eng.get_delta(time_a, time_b)
	assert delta0[1]["units"] == {2: {3: False, 4: False}}
	assert delta0[2]["units"] == {1: {6: True, 7: True}}
	delta1 = eng.get_delta(time_b, time_a)
	assert delta1[1]["units"] == {2: {3: True, 4: True}}
	assert delta1[2]["units"] == {1: {6: False, 7: False}}


def test_character_stat_delta(null_engine, codepath):
	eng = null_engine
	one = eng.new_character(1)
	two = eng.new_character(2)
	one.stat[3] = 4
	one.stat[5] = 6
	one.stat[7] = 8
	two.stat[11] = 12
	time_a = tuple(eng.time)
	if codepath == "branch-delta":
		eng.next_turn()
	else:
		eng.branch = "branch"
	del one.stat[3]
	del one.stat[5]
	two.stat[9] = 10
	time_b = tuple(eng.time)
	delta0 = eng.get_delta(time_a, time_b)
	assert delta0[1][3] is ...
	assert delta0[1][5] is ...
	assert 7 not in delta0[1]
	assert delta0[2][9] == 10
	assert 11 not in delta0[2]
	delta1 = eng.get_delta(time_b, time_a)
	assert delta1[1][3] == 4
	assert delta1[1][5] == 6
	assert 7 not in delta1[1]
	assert delta1[2][9] is ...
	assert 11 not in delta1[2]


def test_node_existence_delta(null_engine, codepath):
	eng = null_engine
	one = eng.new_character(1)
	two = one.new_place(2)
	two.add_thing(3)
	one.add_place(4)
	one.add_thing(5, 4)
	six = eng.new_character(6)
	six.add_place(7)
	six.add_thing(8, 7)
	time_a = tuple(eng.time)
	if codepath == "slow-delta":
		eng.branch = "branch"
	else:
		eng.next_turn()
	del one.place[2]
	six.add_place(9)
	six.add_thing(10, 9)
	time_b = tuple(eng.time)
	delta0 = eng.get_delta(time_a, time_b)
	assert delta0[1]["nodes"] == {3: False, 2: False}
	assert delta0[6]["nodes"] == {9: True, 10: True}
	delta1 = eng.get_delta(time_b, time_a)
	assert delta1[1]["nodes"] == {3: True, 2: True}
	assert delta1[6]["nodes"] == {9: False, 10: False}


def test_node_stat_delta(null_engine, codepath):
	eng = null_engine
	ch = eng.new_character("me")
	one = ch.new_place(1)
	two = ch.new_place(2)
	one[3] = 4
	two[5] = 6
	two[11] = 0
	one[7] = 8
	one[9] = 10
	time_a = tuple(eng.time)
	if codepath == "slow-delta":
		eng.branch = "branch"
	else:
		eng.next_turn()
	del one[7]
	del one[9]
	two[11] = 12
	two[13] = 14
	time_b = tuple(eng.time)
	delta0 = eng.get_delta(time_a, time_b)
	assert delta0["me"]["node_val"] == {
		1: {7: ..., 9: ...},
		2: {11: 12, 13: 14},
	}
	delta1 = eng.get_delta(time_b, time_a)
	assert delta1["me"]["node_val"] == {1: {7: 8, 9: 10}, 2: {11: 0, 13: ...}}


def test_portal_existence_delta(null_engine, codepath):
	pass


def test_thing_location_delta(null_engine, codepath):
	pass


def test_character_rulebook_delta(null_engine, codepath):
	pass


def test_unit_rulebook_delta():
	pass


def test_character_thing_rulebook_delta():
	pass


def test_character_place_rulebook_delta():
	pass


def test_character_portal_rulebook_delta():
	pass


def test_node_rulebook_delta():
	pass


def test_portal_rulebook_delta():
	pass


def test_character_created_delta():
	"""Test whether a delta includes the initial keyframe for characters created"""
	pass



# This file is part of allegedb, an object-relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""WindowDict, the core data structure used by allegedb's caching system.

It resembles a dictionary, more specifically a defaultdict-like where retrieving
a key that isn't set will get the highest set key that is lower than the key
you asked for (and thus, keys must be orderable). It is optimized for retrieval
of the same key and neighboring ones repeatedly and in sequence.

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import (
	ItemsView,
	KeysView,
	Mapping,
	MutableMapping,
	ValuesView,
)
from enum import Enum
from functools import partial
from itertools import chain
from operator import ge, itemgetter, le
from threading import RLock
from typing import Any, Callable, Iterable, Iterator, Union

from .exc import HistoricKeyError
from .types import Tick, Turn, Value

get0 = itemgetter(0)
get1 = itemgetter(1)


class Direction(Enum):
	FORWARD = "forward"
	BACKWARD = "backward"


def update_window(
	turn_from: Turn,
	tick_from: Tick,
	turn_to: Turn,
	tick_to: Tick,
	updfun: Callable[[Turn, Tick, ...], None],
	branchd: SettingsTurnDict,
):
	"""Iterate over some time in ``branchd``, call ``updfun`` on the values"""
	if turn_from == turn_to:
		if turn_from not in branchd:
			return
		for tick, state in (
			branchd[turn_from]
			.future(tick_from, include_same_rev=False)
			.items()
		):
			if tick > tick_to:
				return
			updfun(turn_from, tick, *state)
		return
	if turn_from in branchd:
		for tick, state in (
			branchd[turn_from]
			.future(tick_from, include_same_rev=False)
			.items()
		):
			updfun(turn_from, tick, *state)
	midturn: Turn
	for midturn in range(turn_from + 1, turn_to):
		if midturn in branchd:
			for tick, state in branchd[midturn].items():
				updfun(midturn, tick, *state)
	if turn_to in branchd:
		for tick, state in branchd[turn_to].items():
			if tick > tick_to:
				return
			updfun(turn_to, tick, *state)


def update_backward_window(
	turn_from: Turn,
	tick_from: Tick,
	turn_to: Turn,
	tick_to: Tick,
	updfun: Callable[[Turn, Tick, ...], None],
	branchd: SettingsTurnDict,
):
	"""Iterate backward over time in ``branchd``, call ``updfun`` on the values"""
	if turn_from == turn_to:
		if turn_from not in branchd:
			return
		for tick, state in (
			branchd[turn_from].past(tick_from, include_same_rev=True).items()
		):
			if tick <= tick_to:
				return
			updfun(turn_from, tick, *state)
		return
	if turn_from in branchd:
		for tick, state in (
			branchd[turn_from].past(tick_from, include_same_rev=True).items()
		):
			updfun(turn_from, tick, *state)
	midturn: Turn
	for midturn in range(turn_from - 1, turn_to, -1):
		if midturn in branchd:
			for tick, state in reversed(branchd[midturn].items()):
				updfun(midturn, tick, *state)
	if turn_to in branchd:
		for tick, state in reversed(
			branchd[turn_to].future(tick_to, include_same_rev=False).items()
		):
			updfun(turn_to, tick, *state)


class WindowDictKeysView(KeysView):
	"""Look through all the keys a WindowDict contains."""

	_mapping: "WindowDict"

	def __contains__(self, rev: int):
		with self._mapping._lock:
			return rev in self._mapping._keys

	def __iter__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if past:
				yield from map(get0, past)
			if future:
				yield from map(get0, reversed(future))

	def __reversed__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if future:
				yield from map(get0, future)
			if past:
				yield from map(get0, reversed(past))

	def __repr__(self):
		return f"<WindowDictKeysView containing {list(self)}>"


class WindowDictItemsView(ItemsView):
	"""Look through everything a WindowDict contains."""

	_mapping: "WindowDict"

	def __contains__(self, item: tuple[int, Any]):
		with self._mapping._lock:
			return item in self._mapping._past or item in self._mapping._future

	def __iter__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if past:
				yield from past
			if future:
				yield from reversed(future)

	def __reversed__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if future:
				yield from future
			if past:
				yield from reversed(past)


class WindowDictPastKeysView(KeysView):
	"""View on a WindowDict's keys relative to last lookup"""

	_mapping: WindowDictPastView

	def __iter__(self):
		with self._mapping.lock:
			yield from map(get0, reversed(self._mapping.stack))

	def __reversed__(self):
		with self._mapping.lock:
			yield from map(get0, self._mapping.stack)

	def __contains__(self, item: int):
		return item in self._mapping


class WindowDictFutureKeysView(KeysView):
	_mapping: WindowDictFutureView

	def __iter__(self):
		with self._mapping.lock:
			yield from map(get0, self._mapping.stack)

	def __reversed__(self):
		with self._mapping.lock:
			yield from map(get0, reversed(self._mapping.stack))

	def __contains__(self, item):
		return item in self._mapping


class WindowDictPastFutureItemsView(ItemsView):
	_mapping: Union["WindowDictPastView", "WindowDictFutureView"]

	@staticmethod
	@abstractmethod
	def _out_of_range(item: tuple, stack: list):
		pass

	def __iter__(self):
		with self._mapping.lock:
			yield from reversed(self._mapping.stack)

	def __reversed__(self):
		with self._mapping.lock:
			yield from self._mapping.stack

	def __contains__(self, item: tuple[int, Any]):
		with self._mapping.lock:
			if self._out_of_range(item, self._mapping.stack):
				return False
			k, v = item
			return self._mapping[k] == v


class WindowDictPastItemsView(WindowDictPastFutureItemsView):
	@staticmethod
	def _out_of_range(item: tuple[int, Any], stack: list[tuple[int, Any]]):
		return item[0] < stack[0][0] or item[0] > stack[-1][0]


class WindowDictFutureItemsView(WindowDictPastFutureItemsView):
	"""View on a WindowDict's future items relative to last lookup"""

	@staticmethod
	def _out_of_range(item: tuple[int, Any], stack: list[tuple[int, Any]]):
		return item[0] < stack[-1][0] or item[0] > stack[0][0]


class WindowDictPastFutureValuesView(ValuesView):
	"""Abstract class for views on the past or future values of a WindowDict"""

	_mapping: Union["WindowDictPastView", "WindowDictFutureView"]

	def __iter__(self):
		with self._mapping.lock:
			yield from map(get1, reversed(self._mapping.stack))

	def __contains__(self, item: Any):
		with self._mapping.lock:
			return item in map(get1, self._mapping.stack)


class WindowDictValuesView(ValuesView):
	"""Look through all the values that a WindowDict contains."""

	_mapping: "WindowDict"

	def __contains__(self, value: Any):
		with self._mapping._lock:
			return value in map(get1, self._mapping._past) or value in map(
				get1, self._mapping._future
			)

	def __iter__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if past:
				yield from map(get1, past)
			if future:
				yield from map(get1, reversed(future))

	def __reversed__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if future:
				yield from map(get1, future)
			if past:
				yield from map(get1, reversed(past))


class WindowDictPastFutureView(ABC, Mapping):
	"""Abstract class for historical views on WindowDict"""

	__slots__ = ("stack", "lock")
	stack: list[tuple[int, Value]]

	def __init__(self, stack: list[tuple[int, Value]], lock: RLock) -> None:
		self.stack = stack
		self.lock = lock

	def __len__(self) -> int:
		with self.lock:
			stack = self.stack
			if not stack:
				return 0
			return len(stack)

	def __copy__(self):
		return type(self)(self.stack.copy(), RLock())


class WindowDictPastView(WindowDictPastFutureView):
	"""Read-only mapping of just the past of a WindowDict

	Iterates in descending order

	"""

	def __iter__(self) -> Iterator[int]:
		with self.lock:
			yield from map(get0, reversed(self.stack))

	def __reversed__(self) -> Iterator[int]:
		with self.lock:
			yield from map(get0, self.stack)

	def __getitem__(self, key: int) -> Value:
		with self.lock:
			stack = self.stack
			if not stack or key < stack[0][0] or key > stack[-1][0]:
				raise KeyError("Out of range", key)
			return _recurse(key, stack)[1]

	def keys(self) -> WindowDictPastKeysView:
		return WindowDictPastKeysView(self)

	def items(self) -> WindowDictPastItemsView:
		return WindowDictPastItemsView(self)

	def values(self) -> WindowDictPastFutureValuesView:
		return WindowDictPastFutureValuesView(self)


class WindowDictFutureView(WindowDictPastFutureView):
	"""Read-only mapping of just the future of a WindowDict

	Iterates in ascending order

	"""

	def __iter__(self) -> Iterator[int]:
		with self.lock:
			yield from map(get0, reversed(self.stack))

	def __reversed__(self) -> Iterator[int]:
		with self.lock:
			yield from map(get0, self.stack)

	def __getitem__(self, key: int):
		with self.lock:
			stack = list(reversed(self.stack))
			if not stack:
				raise KeyError("No data")
			if key < stack[0][0] or key > stack[-1][0]:
				raise KeyError("No such revision", key)
			return _recurse(key, stack)[1]

	def keys(self) -> WindowDictFutureKeysView:
		return WindowDictFutureKeysView(self)

	def items(self) -> WindowDictFutureItemsView:
		return WindowDictFutureItemsView(self)

	def values(self) -> WindowDictPastFutureValuesView:
		return WindowDictPastFutureValuesView(self)


class WindowDictSlice:
	__slots__ = ["dic", "slic"]
	dic: "WindowDict"
	slic: slice

	def __init__(self, dic: "WindowDict", slic: slice):
		self.dic = dic
		self.slic = slic

	def __iter__(self) -> Iterator[Value]:
		with self.dic._lock:
			slic = self.slic
			if slic.step:
				return self._step_iter()
			start, stop = self._modulo(slic.start, slic.stop)
			if stop < start:
				it = self._get_item_iterator_stepless_reversed(
					stop, start, include_start=False, include_stop=True
				)
			else:
				it = self._get_item_iterator_stepless(start, stop)
			return map(get1, it)

	def __reversed__(self) -> Iterator[Value]:
		with self.dic._lock:
			slic = self.slic
			if slic.step:
				return self._step_iter_reversed()
			start, stop = self._modulo(slic.start, slic.stop)
			if stop < start:
				items = self._get_item_iterator_stepless(
					stop, start, include_start=False, include_stop=True
				)
			else:
				items = self._get_item_iterator_stepless_reversed(start, stop)
			return map(get1, items)

	def _step_iter(self):
		dic = self.dic
		slic = self.slic
		start, stop = self._modulo(slic.start, slic.stop)
		for i in range(start, stop, slic.step):
			yield dic[i]

	def _step_iter_reversed(self):
		dic = self.dic
		slic = self.slic
		start, stop = self._modulo(slic.start, slic.stop)
		for i in reversed(range(start, stop, slic.step)):
			yield dic[i]

	@staticmethod
	def _iter_items_until(
		it: Iterator[tuple[int, Value]],
		until: Callable[[int], bool],
		include_last: bool = False,
	) -> Iterator[tuple[int, Value]]:
		if include_last:
			for rev, val in it:
				yield rev, val
				if until(rev):
					return
		else:
			for rev, val in it:
				if until(rev):
					return
				yield rev, val

	def _iter_past_items_until(
		self,
		rev: int,
		until: Callable[[int], bool],
		include_last: bool = False,
	):
		return self._iter_items_until(
			iter(self.dic.past(rev).items()), until, include_last
		)

	def _iter_future_items_until(
		self,
		rev: int,
		until: Callable[[int], bool],
		include_last: bool = False,
	):
		return self._iter_items_until(
			iter(self.dic.future(rev).items()), until, include_last
		)

	def _modulo(self, start: int | None, stop: int | None) -> tuple[int, int]:
		dic = self.dic
		biggest = 0
		if dic._future:
			biggest = dic._future[0][0]
		elif dic._past:
			biggest = dic._past[-1][0]
		if start is None:
			start = 0
		elif start < 0:
			start = biggest + start
			if start < 0:
				raise IndexError("WindowDict index out of range", start)
		if stop is None:
			stop = biggest + 1
		elif stop < 0:
			stop = biggest + stop
			if stop < 0:
				raise IndexError("WindowDict index out of range", stop)
		return start, stop

	def _get_item_iterator_stepless_reversed(
		self,
		start: int,
		stop: int,
		include_start: bool = True,
		include_stop: bool = False,
	) -> Iterator[tuple[int, Value]]:
		if isinstance(start, int) and isinstance(stop, int) and start > stop:
			raise ValueError("start should come before stop", start, stop)
		dic = self.dic
		if not dic:
			return iter(())
		seek = dic._seek
		past = dic._past
		future = dic._future
		if not past and not future:
			return iter(())
		if stop == start:
			seek(start)
			if past and past[-1][0] == start:
				return iter((past[-1][1],))
			else:
				return iter(())
		seek(stop)
		if not include_stop and past and past[-1][0] == stop:
			future.append(past.pop())
		return self._iter_items_until(
			reversed(past), partial(ge, start), include_last=include_start
		)

	def _get_item_iterator_stepless(
		self,
		start: int,
		stop: int,
		include_start: bool = True,
		include_stop: bool = False,
	) -> Iterator[tuple[int, Value]]:
		if isinstance(start, int) and isinstance(stop, int) and start > stop:
			raise ValueError("start should come before stop", start, stop)
		dic = self.dic
		if not dic:
			return iter(())
		seek = dic._seek
		past = dic._past
		future = dic._future
		if not past and not future:
			return iter(())
		if stop == start:
			seek(start)
			if past and past[-1][0] == start:
				return iter((past[-1][1],))
			else:
				return iter(())
		it = self._iter_future_items_until(
			start, partial(le, stop), include_last=include_stop
		)
		if include_start and past and past[-1][0] == start:
			return chain((past[-1],), it)
		return it


def _recurse(rev: int, revs: list[tuple[int, Any]]) -> tuple[int, Any]:
	if len(revs) < 1:
		raise HistoricKeyError("No data ever for revision", rev, deleted=False)
	elif len(revs) == 1:
		if revs[0][0] <= rev:
			return revs[0]
		raise HistoricKeyError("Can't retrieve revision", rev, deleted=False)
	pivot = len(revs) // 2
	before = revs[:pivot]
	after = revs[pivot:]
	assert before and after
	if rev < after[0][0]:
		if rev > before[-1][0]:
			return before[-1]
		return _recurse(rev, before)
	elif rev == after[0][0]:
		return after[0]
	else:
		return _recurse(rev, after)


class WindowDict(MutableMapping):
	"""A dict that keeps every value that a variable has had over time.

	Look up a revision number in this dict, and it will give you the
	effective value as of that revision. Keys should always be
	revision numbers.

	Optimized for the cases where you look up the same revision
	repeatedly, or its neighbors.

	This supports slice notation to get all values in a given
	time-frame. If you do not supply a step, you'll just get the
	values, with no indication of when they're from exactly --
	so explicitly supply a step of 1 to get the value at each point in
	the slice, or use the ``future`` and ``past`` methods to get read-only
	mappings of data relative to a particular revision.

	Unlike slices of eg. lists, you can slice with a start greater than the stop
	even if you don't supply a step. That will get you values in reverse order.

	"""

	__slots__ = ("_future", "_past", "_keys", "_lock")

	_past: list[tuple[int, Any]]
	_future: list[tuple[int, Any]]
	_keys: set[int]

	@property
	def beginning(self) -> int | None:
		with self._lock:
			if not self._past:
				if not self._future:
					return None
				return self._future[-1][0]
			return self._past[0][0]

	@property
	def end(self) -> int | None:
		with self._lock:
			if not self._future:
				if not self._past:
					return None
				return self._past[-1][0]
			return self._future[0][0]

	def future(
		self, rev: int, include_same_rev: bool = False, copy: bool = False
	) -> WindowDictFutureView:
		"""Return a Mapping of items after the given revision.

		:param include_same_rev: Whether to include the specified revision in
			the Mapping, if present in the `WindowDict`. `False` by default.
		:param copy: Whether to make a copy of the data, so that the Mapping
			won't mutate when you access the underlying `WindowDict`. Default
			`False`. The Mapping has a `copy` method, too.

		"""
		if not isinstance(rev, int):
			raise TypeError("rev must be int")
		with self._lock:
			self._seek(rev)
			if include_same_rev and self._past and self._past[-1][0] == rev:
				self._future.append(self._past.pop())
			if copy:
				future = self._future.copy()
				lock = self._lock
			else:
				future = self._future
				lock = RLock()
		return WindowDictFutureView(future, lock)

	def past(
		self, rev: int, include_same_rev: bool = True, copy: bool = False
	) -> WindowDictPastView:
		"""Return a Mapping of items at or before the given revision.

		:param include_same_rev: Whether to include the specified revision in
			the Mapping, if present in the `WindowDict`. `True` by default.
		:param copy: Whether to make a copy of the data, so that the Mapping
			won't mutate when you access the underlying `WindowDict`. Default
			`False`. The Mapping has a `copy` method, too.

		"""
		if not isinstance(rev, int):
			raise TypeError("rev must be int")
		with self._lock:
			self._seek(rev)
			if (
				not include_same_rev
				and self._future
				and self._future[-1][0] == rev
			):
				self._past.append(self._future.pop())
			if copy:
				past = self._past.copy()
				lock = RLock()
			else:
				past = self._past
				lock = self._lock
		return WindowDictPastView(past, lock)

	def search(self, rev: int) -> Any:
		"""Alternative access for far-away revisions

		This uses a binary search, which is faster in the case of random
		access, but not in the case of fast-forward and rewind, which are
		more common in time travel.

		This doesn't change the state of the cache.

		"""

		with self._lock:
			revs = self._past + list(reversed(self._future))
			if len(revs) == 1:
				result_rev, result = revs[0]
				if rev < result_rev:
					raise HistoricKeyError(
						"No data ever for revision", rev, deleted=False
					)
			else:
				result_rev, result = _recurse(rev, revs)
			return result

	def _seek(self, rev: int) -> None:
		"""Arrange the caches to help look up the given revision."""
		if not isinstance(rev, int):
			raise TypeError("Need integer revision", rev)
		past = self._past
		future = self._future
		if future:
			appender = past.append
			popper = future.pop
			future_start = future[-1][0]
			while future_start <= rev:
				appender(popper())
				if future:
					future_start = future[-1][0]
				else:
					break
		if past:
			popper = past.pop
			appender = future.append
			past_end = past[-1][0]
			while past_end > rev:
				appender(popper())
				if past:
					past_end = past[-1][0]
				else:
					break

	def rev_gettable(self, rev: int) -> bool:
		beg = self.beginning
		if beg is None:
			return False
		return rev >= beg

	def rev_before(self, rev: int, search=False) -> int | None:
		"""Return the latest past rev on which the value changed.

		If it changed on this exact rev, return the rev.

		"""
		with self._lock:
			if search:
				rev, _ = _recurse(
					rev, self._past + list(reversed(self._future))
				)
				return rev
			else:
				self._seek(rev)
				if self._past:
					return self._past[-1][0]
				else:
					return None

	def rev_after(self, rev: int) -> int | None:
		"""Return the earliest future rev on which the value will change."""
		with self._lock:
			self._seek(rev)
			if self._future:
				return self._future[-1][0]
			else:
				return None

	def initial(self) -> Any:
		"""Return the earliest value we have"""
		with self._lock:
			if self._past:
				return self._past[0][1]
			if self._future:
				return self._future[-1][1]
			raise KeyError("No data")

	def final(self) -> Any:
		"""Return the latest value we have"""
		with self._lock:
			if self._future:
				return self._future[0][1]
			if self._past:
				return self._past[-1][1]
			raise KeyError("No data")

	def truncate(
		self, rev: int, direction: Direction = Direction.FORWARD
	) -> set[int]:
		"""Delete everything after the given revision, exclusive.

		With direction='backward', delete everything before the revision,
		exclusive, instead.

		Return a set of keys deleted.

		"""
		if not isinstance(direction, Direction):
			direction = Direction(direction)
		deleted = set()
		with self._lock:
			self._seek(rev)
			if direction == Direction.FORWARD:
				to_delete = set(map(get0, self._future))
				deleted.update(to_delete)
				self._keys.difference_update(to_delete)
				self._future = []
			elif direction == Direction.BACKWARD:
				if not self._past:
					return deleted
				if self._past[-1][0] == rev:
					to_delete = set(map(get0, self._past[:-1]))
					deleted.update(to_delete)
					self._keys.difference_update(to_delete)
					self._past = [self._past[-1]]
				else:
					to_delete = set(map(get0, self._past))
					deleted.update(to_delete)
					self._keys.difference_update(to_delete)
					self._past = []
			else:
				raise ValueError("Need direction 'forward' or 'backward'")
		return deleted

	def keys(self) -> WindowDictKeysView:
		return WindowDictKeysView(self)

	def items(self) -> WindowDictItemsView:
		return WindowDictItemsView(self)

	def values(self) -> WindowDictValuesView:
		return WindowDictValuesView(self)

	def __bool__(self) -> bool:
		return bool(self._keys)

	def copy(self):
		with self._lock:
			empty = WindowDict.__new__(WindowDict)
			empty._past = self._past.copy()
			empty._future = self._future.copy()
			empty._keys = self._keys.copy()
			return empty

	def __init__(
		self, data: Union[list[tuple[int, Any]], dict[int, Any]] = None
	) -> None:
		self._lock = RLock()
		with self._lock:
			if not data:
				self._past = []
			elif isinstance(data, Mapping):
				self._past = list(data.items())
			else:
				# assume it's an orderable sequence of pairs
				self._past = list(data)
			self._past.sort()
			self._future = []
			self._keys = set(map(get0, self._past))

	def __iter__(self) -> Iterable[Any]:
		if not self:
			return
		with self._lock:
			if self._past:
				yield from map(get0, self._past)
			if self._future:
				yield from map(get0, self._future)

	def __contains__(self, item: int) -> bool:
		with self._lock:
			return item in self._keys

	def __len__(self) -> int:
		with self._lock:
			return len(self._keys)

	def __getitem__(self, rev: int) -> Any:
		if isinstance(rev, slice):
			return WindowDictSlice(self, rev)
		with self._lock:
			self._seek(rev)
			past = self._past
			if not past:
				raise HistoricKeyError(
					"Revision {} is before the start of history".format(rev)
				)
			return past[-1][1]

	def __setitem__(self, rev: int, v: Any) -> None:
		past = self._past
		with self._lock:
			if past or self._future:
				self._seek(rev)
				if past:
					if past[-1][0] == rev:
						past[-1] = (rev, v)
					else:
						past.append((rev, v))
				else:
					past.append((rev, v))
			else:
				past.append((rev, v))
			self._keys.add(rev)

	def __delitem__(self, rev: int) -> None:
		# Not checking for rev's presence at the beginning because
		# to do so would likely require iterating thru history,
		# which I have to do anyway in deleting.
		# But handle degenerate case.
		if not self:
			raise HistoricKeyError("Tried to delete from an empty WindowDict")
		if self.beginning is None:
			if self.end is not None and rev > self.end:
				raise HistoricKeyError(
					"Rev outside of history: {}".format(rev)
				)
		elif self.end is None:
			if self.beginning is not None and rev < self.beginning:
				raise HistoricKeyError(
					"Rev outside of history: {}".format(rev)
				)
		elif not self.beginning <= rev <= self.end:
			raise HistoricKeyError("Rev outside of history: {}".format(rev))
		with self._lock:
			self._seek(rev)
			past = self._past
			if not past or past[-1][0] != rev:
				raise HistoricKeyError("Rev not present: {}".format(rev))
			del past[-1]
			self._keys.remove(rev)

	def __repr__(self) -> str:
		with self._lock:
			me = {}
			if self._past:
				me.update(self._past)
			if self._future:
				me.update(self._future)
			return "{}({})".format(self.__class__.__name__, me)


class FuturistWindowDict(WindowDict):
	"""A WindowDict that does not let you rewrite the past."""

	__slots__ = (
		"_future",
		"_past",
	)
	_future: list[tuple[int, Any]]
	_past: list[tuple[int, Any]]

	def __setitem__(self, rev: int, v: Any) -> None:
		if hasattr(v, "unwrap") and not hasattr(v, "no_unwrap"):
			v = v.unwrap()
		with self._lock:
			self._seek(rev)
			past = self._past
			future = self._future
			if future:
				raise HistoricKeyError(
					"Already have some history after {}".format(rev)
				)
			if not past:
				past.append((rev, v))
			elif rev > past[-1][0]:
				past.append((rev, v))
			elif rev == past[-1][0]:
				past[-1] = (rev, v)
			else:
				raise HistoricKeyError(
					"Already have some history after {} "
					"(and my seek function is broken?)".format(rev)
				)
			self._keys.add(rev)


class TurnDict(FuturistWindowDict):
	__slots__ = ("_future", "_past")
	_future: list[tuple[Turn, Any]]
	_past: list[tuple[Turn, Any]]
	cls = FuturistWindowDict

	def __setitem__(self, turn: Turn, value: Any) -> None:
		if type(value) is not FuturistWindowDict:
			value = FuturistWindowDict(value)
		FuturistWindowDict.__setitem__(self, turn, value)


class EntikeyWindowDict(WindowDict):
	__slots__ = ("_past", "_future", "entikeys")

	def __init__(
		self, data: Union[list[tuple[int, Any]], dict[int, Any]] = None
	) -> None:
		if data:
			if hasattr(data, "values") and callable(data.values):
				self.entikeys = {value[:-2] for value in data.values()}
			else:
				self.entikeys = {value[:-2] for value in data}
		else:
			self.entikeys = set()
		super().__init__(data)

	def __setitem__(self, rev: int, v: tuple) -> None:
		self.entikeys.add(v[:-2])
		super().__setitem__(rev, v)

	def __delitem__(self, rev: int) -> None:
		entikey = self[rev][:-2]
		super().__delitem__(rev)
		for tup in self.values():
			if tup[:-2] == entikey:
				return
		self.entikeys.remove(entikey)


class SettingsTurnDict(WindowDict):
	"""A WindowDict that contains a span of time, indexed as turns and ticks

	Each turn is a series of ticks. Once a value is set at some turn and tick,
	it's in effect at every tick in the turn after that one, and every
	further turn.

	"""

	__slots__ = ("_future", "_past")
	_future: list[tuple[Turn, Any]]
	_past: list[tuple[Turn, Any]]
	cls = WindowDict

	def __setitem__(self, turn: Turn, value: cls | dict) -> None:
		if not isinstance(value, self.cls):
			value = self.cls(value)
		WindowDict.__setitem__(self, turn, value)

	def retrieve(self, turn: Turn, tick: Tick) -> Any:
		"""Retrieve the value that was in effect at this turn and tick

		Whether or not it was *set* at this turn and tick

		"""
		if turn in self and self[turn].rev_gettable(tick):
			return self[turn][tick]
		elif self.rev_gettable(turn - 1):
			return self[turn - 1].final()
		raise KeyError(f"Can't retrieve turn {turn}, tick {tick}")

	def retrieve_exact(self, turn: Turn, tick: Tick) -> Any:
		"""Retrieve the value only if it was set at this exact turn and tick"""
		if turn not in self:
			raise KeyError(f"No data in turn {turn}")
		if tick not in self[turn]:
			raise KeyError(f"No data for tick {tick} in turn {turn}")
		return self[turn][tick]

	def store_at(self, turn: Turn, tick: Tick, value: Any) -> None:
		"""Set a value at a time, creating the turn if needed"""
		if turn in self:
			self[turn][tick] = value
		else:
			self[turn] = {tick: value}


class EntikeySettingsTurnDict(SettingsTurnDict):
	cls = EntikeyWindowDict



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import tempfile
from shutil import rmtree

import pytest

from lisien.engine import Engine
from lisien.tests.data import CHARACTER_UPDATES


def set_in_mapping(mapp, stat, v):
	"""Sync a value in ``mapp``, having key ``stat``, with ``v``."""
	# Mutate the stuff in-place instead of simply replacing it,
	# because this could trigger side effects
	if stat == "name":
		return
	if v is ...:
		del mapp[stat]
		return
	if stat not in mapp:
		mapp[stat] = v
		return
	if isinstance(v, (dict, set)):
		mapp[stat].update(v)
		for item in list(mapp[stat]):
			if item not in v:
				try:
					del mapp[stat][item]
				except TypeError:
					mapp[stat].remove(item)
	elif isinstance(v, list):
		for item in list(mapp[stat]):
			if item not in v:
				mapp[stat].remove(item)
		for i, item in enumerate(v):
			if mapp[stat][i] != item:
				mapp[stat].insert(i, item)
	else:
		mapp[stat] = v


def update_char(char, *, stat=(), nodes=(), portals=()):
	"""Make a bunch of changes to a character-like object"""

	def update(d, dd):
		for k, v in dd.items():
			if v is ... and k in d:
				del d[k]
			else:
				d[k] = v

	end_stats = dict(char.stat)
	for stat, v in stat:
		set_in_mapping(char.stat, stat, v)
		if v is ... and stat in end_stats:
			del end_stats[stat]
		else:
			end_stats[stat] = v
	end_places = dict(char.place)
	end_things = dict(char.thing)
	for node, v in nodes:
		if v is ...:
			del char.node[node]
			if node in end_places:
				del end_places[node]
			if node in end_things:
				del end_things[node]
		elif node in char.place:
			if "location" in v:
				del end_places[node]
				char.place2thing(node, v.pop("location"))
				if node in end_places:
					me = end_things[node] = end_places.pop(node)
				else:
					me = end_things[node] = dict(char.thing[node])
				update(me, v)
				for k, vv in v.items():
					set_in_mapping(char.thing[node], k, vv)
			else:
				if node in end_places:
					me = end_places[node]
				else:
					me = end_places[node] = dict(char.place[node])
				update(me, v)
				for k, vv in v.items():
					set_in_mapping(char.place[node], k, vv)
		elif node in char.thing:
			if "location" in v and v["location"] is None:
				if node in end_things:
					me = end_places[node] = end_things.pop(node)
				else:
					me = end_places[node] = dict(char.thing[node])
				del me["location"]
				del v["location"]
				char.thing2place(node)
				del v["location"]
				for k, vv in v.items():
					set_in_mapping(char.place[node], k, vv)
					set_in_mapping(end_places[node], k, vv)
			else:
				for k, vv in v.items():
					set_in_mapping(char.thing[node], k, vv)
					set_in_mapping(end_things[node], k, vv)
		elif "location" in v:
			end_things[node] = v
			me = char.new_thing(node, v.pop("location"))
			for k, vv in v.items():
				set_in_mapping(me, k, vv)
		else:
			end_places[node] = v
			me = char.new_node(node)
			for k, vv in v.items():
				set_in_mapping(me, k, vv)
	end_edges = {}
	for orig, dests in char.portal.items():
		end_edges[orig] = here = {}
		for dest, port in dests.items():
			here[dest] = dict(port)
	for o, d, v in portals:
		if v is ...:
			del char.edge[o][d]
			del end_edges[o][d]
		else:
			me = end_edges.setdefault(o, {}).setdefault(d, {})
			update(me, v)
			e = char.new_portal(o, d)
			for k, vv in v.items():
				set_in_mapping(e, k, vv)
	return {
		"stat": end_stats,
		"place": end_places,
		"thing": end_things,
		"portal": end_edges,
	}


@pytest.mark.parametrize(
	["name", "data", "stat", "nodestat", "statup", "nodeup", "edgeup"],
	CHARACTER_UPDATES,
)
def test_char_creation(
	engy, name, data, stat, nodestat, statup, nodeup, edgeup
):
	char = engy.new_character(name, data, **stat)
	assert set(char.node) == set(data)
	es = set()
	for k, v in data.items():
		for vv in v:
			es.add((k, vv))
	assert set(char.edges) == es
	assert char.stat == stat


@pytest.fixture(params=CHARACTER_UPDATES)
def char_data(request):
	return request.param


def test_facade_creation(tmp_path, char_data):
	name, data, stat, nodestat, statup, nodeup, edgeup = char_data
	with Engine(tmp_path, workers=0) as eng:
		char = eng.new_character(name, data, **stat)
		fac = char.facade()
		assert dict(fac.node) == dict(char.node)
		assert fac.node == char.node
		assert fac.edges == char.edges
		assert set(fac.edges) == set(char.edges)
		assert fac.stat == char.stat
		assert dict(fac.stat) == dict(char.stat)


# TODO parametrize bunch of characters
@pytest.fixture(scope="function", params=CHARACTER_UPDATES)
def character_updates(request, sqleng):
	name, data, stat, nodestat, statup, nodeup, edgeup = request.param
	char = sqleng.new_character(name, data, **stat)
	update_char(char, nodes=nodestat)
	yield char, statup, nodeup, edgeup


def test_facade(character_updates):
	"""Make sure you can alter a facade independent of the character it's from"""
	character, statup, nodeup, edgeup = character_updates
	start_stat = dict(character.stat)
	start_place = dict(character.place)
	start_thing = dict(character.thing)
	start_edge = {}
	for o in character.edge:
		for d in character.edge[o]:
			start_edge.setdefault(o, {})[d] = dict(
				character.edge[o][d].items()
			)
	facade = character.facade()
	updated = update_char(facade, stat=statup, nodes=nodeup, portals=edgeup)
	assert facade.stat == updated["stat"]
	assert facade.place == updated["place"]
	assert facade.thing == updated["thing"]
	assert facade.portal == updated["portal"]
	# changes to a facade should not impact the underlying character
	assert start_stat == dict(character.stat)
	assert start_place == dict(character.place)
	assert start_thing == dict(character.thing)
	end_edge = {}
	for o in character.edge:
		for d in character.edge[o]:
			end_edge.setdefault(o, {})[d] = dict(character.edge[o][d])
	assert start_edge == end_edge


def test_set_rulebook(engine):
	eng = engine
	eng.universal["list"] = []
	ch = eng.new_character("physical")

	@ch.rule(always=True)
	def rule0(cha):
		cha.engine.universal["list"].append(0)

	@eng.rule(always=True)
	def rule1(who):
		who.engine.universal["list"].append(1)

	eng.rulebook["rb1"] = [rule1]
	eng.next_turn()
	assert eng.universal["list"] == [0]
	default_rulebook_name = ch.rulebook.name
	ch.rulebook = "rb1"
	eng.next_turn()
	assert eng.universal["list"] == [0, 1]
	ch.rulebook = eng.rulebook[default_rulebook_name]
	eng.next_turn()
	assert eng.universal["list"] == [0, 1, 0]


def test_iter_portals(sqleng):
	from lisien.character import grid_2d_8graph

	ch = sqleng.new_character("physical", grid_2d_8graph(4, 4))
	portal_abs = {
		(portal.origin.name, portal.destination.name)
		for portal in ch.portals()
	}
	for a, ayes in ch.adj.items():
		for b in ayes:
			assert (a, b) in portal_abs
	for a, b in portal_abs:
		assert a in ch.edge
		assert b in ch.edge[a]



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Common classes for collections in lisien

Notably includes wrappers for mutable objects, allowing them to be stored in
the database. These simply store the new value.

Most of these are subclasses of :class:`blinker.Signal`, so you can listen
for changes using the ``connect(..)`` method.

"""

from __future__ import annotations

import importlib.util
import json
import os
import sys
from abc import ABC, abstractmethod
from ast import Expr, Module, parse
from collections import UserDict
from collections.abc import MutableMapping
from copy import deepcopy
from inspect import getsource
from io import StringIO
from typing import TYPE_CHECKING

import networkx as nx
from astunparse import Unparser
from blinker import Signal

from .graph import GraphsMapping
from .types import CharName, Key
from .util import AbstractEngine, dedent_source, getatt
from .wrap import wrapval

if TYPE_CHECKING:
	from .character import Character


class TabUnparser(Unparser):
	def fill(self, text: str = ""):
		__doc__ = Unparser.fill.__doc__
		self.f.write("\n" + "\t" * self._indent + text)


def unparse(tree):
	v = StringIO()
	TabUnparser(tree, file=v)
	return v.getvalue()


class AbstractLanguageDescriptor(Signal, ABC):
	@abstractmethod
	def _get_language(self, inst: StringStore) -> str:
		pass

	@abstractmethod
	def _set_language(self, inst: StringStore, val: str) -> None:
		pass

	def __get__(self, instance: StringStore, owner=None):
		return self._get_language(instance)

	def __set__(self, inst: StringStore, val: str):
		self._set_language(inst, val)
		self.send(inst, language=val)


class LanguageDescriptor(AbstractLanguageDescriptor):
	def _get_language(self, inst: StringStore) -> str:
		return inst._current_language

	def _set_language(self, inst, lang):
		if lang != inst._current_language:
			inst._switch_language(lang)
			if (
				not getattr(inst.engine, "_worker", False)
				and inst.engine.eternal["language"] != lang
			):
				inst.engine.eternal["language"] = lang
			inst._current_language = lang


class TamperEvidentDict(dict):
	tampered: bool

	def __init__(self, data=()):
		self.tampered = False
		super().__init__(data)

	def __setitem__(self, key, value):
		self.tampered = True
		super().__setitem__(key, value)

	def __delitem__(self, key):
		self.tampered = True
		super().__delitem__(key)


class ChangeTrackingDict(UserDict):
	def __init__(self, data=()):
		self.changed = {}
		super().__init__(data)

	def apply_changes(self):
		self.data.update(self.changed)
		self.changed.clear()

	def copy(self):
		ret = {}
		ret.update(self.data)
		ret.update(self.changed)
		return ret

	def clear(self):
		self.data.clear()
		self.changed.clear()

	def __contains__(self, item):
		return item in self.changed or item in self.data

	def __iter__(self):
		yield from self.changed
		yield from self.data

	def __len__(self):
		return len(self.changed) + len(self.data)

	def __getitem__(self, item):
		if item in self.changed:
			return self.changed[item]
		return self.data[item]

	def __setitem__(self, key, value):
		self.changed[key] = value

	def __delitem__(self, key):
		if key in self.changed:
			del self.changed[key]
			if key in self.data:
				del self.data[key]
		else:
			del self.data[key]


class StringStore(MutableMapping, Signal):
	language = LanguageDescriptor()

	def __init__(
		self,
		engine_or_string_dict: AbstractEngine | dict,
		prefix: str | None,
		lang="eng",
	):
		super().__init__()
		if isinstance(engine_or_string_dict, dict):
			self._prefix = None
			self._current_language = lang
			if lang in engine_or_string_dict and isinstance(
				engine_or_string_dict[lang], dict
			):
				self._languages = engine_or_string_dict
			else:
				self._languages = {lang: engine_or_string_dict}
		else:
			self.engine = engine_or_string_dict
			self._languages = {lang: TamperEvidentDict()}
			self._prefix = prefix
			self._current_language = lang
			self._switch_language(lang)

	def _switch_language(self, lang):
		"""Write the current language to disk, and load the new one if available"""
		if self._prefix is None:
			if lang not in self._languages:
				self._languages[lang] = {}
			return
		try:
			with open(os.path.join(self._prefix, lang + ".json"), "r") as inf:
				self._languages[lang] = TamperEvidentDict(json.load(inf))
		except FileNotFoundError:
			self._languages[lang] = TamperEvidentDict()
		assert self._current_language in self._languages
		if getattr(self._languages[self._current_language], "tampered", False):
			with open(
				os.path.join(self._prefix, self._current_language + ".json"),
				"w",
			) as outf:
				json.dump(
					self._languages[self._current_language],
					outf,
					indent=4,
					sort_keys=True,
				)
			self._languages[self._current_language].tampered = False

	def __iter__(self):
		return iter(self._languages[self._current_language])

	def __len__(self):
		return len(self._languages[self._current_language])

	def __getitem__(self, k):
		return self._languages[self._current_language][k]

	def __setitem__(self, k, v):
		"""Set the value of a string for the current language."""
		self._languages[self._current_language][k] = v
		self.send(self, key=k, val=v)

	def __delitem__(self, k):
		"""Delete the string from the current language, and remove it from the
		cache.

		"""
		del self._languages[self._current_language][k]
		self.send(self, key=k, val=None)

	def lang_items(self, lang=None):
		"""Yield pairs of (id, string) for the given language."""
		if (
			self._prefix is not None
			and lang is not None
			and self._current_language != lang
		):
			with open(os.path.join(self._prefix, lang + ".json"), "r") as inf:
				self._languages[lang] = TamperEvidentDict(json.load(inf))
		yield from self._languages[lang or self._current_language].items()

	def save(self, reimport=False):
		if self._prefix is None:
			return
		if not os.path.exists(self._prefix):
			os.mkdir(self._prefix)
		for lang, d in self._languages.items():
			if not d.tampered:
				continue
			with open(
				os.path.join(self._prefix, lang + ".json"),
				"w",
			) as outf:
				json.dump(
					self._languages[lang],
					outf,
					indent=4,
					sort_keys=True,
				)
			d.tampered = False
		if reimport:
			with open(
				os.path.join(self._prefix, self._current_language + ".json"),
				"r",
			) as inf:
				self._languages[self._current_language] = TamperEvidentDict(
					json.load(inf)
				)


class FunctionStore(Signal):
	"""A module-like object that lets you alter its code and save your changes.

	Instantiate it with a path to a file that you want to keep the code in.
	Assign functions to its attributes, then call its ``save()`` method,
	and they'll be unparsed and written to the file.

	This is a ``Signal``, so you can pass a function to its ``connect`` method,
	and it will be called when a function is added, changed, or deleted.
	The keyword arguments will be ``attr``, the name of the function, and ``val``,
	the function itself.

	"""

	def __init__(
		self, filename: str | None, initial: dict = None, module: str = None
	):
		if initial is None:
			initial = {}
		super().__init__()
		if filename is None:
			self._filename = None
			self._module = module
			self._ast = Module(body=[])
			self._ast_idx = {}
			self._need_save = False
			self._locl = initial
		else:
			if not filename.endswith(".py"):
				raise ValueError(
					"FunctionStore can only work with pure Python source code"
				)
			self._filename = os.path.abspath(os.path.realpath(filename))
			try:
				self.reimport()
			except (FileNotFoundError, ModuleNotFoundError):
				self._module = module
				self._ast = Module(body=[])
				self._ast_idx = {}
				self.save()
			self._need_save = False
			self._locl = {}
			for k, v in initial.items():
				self[k] = v

	def __dir__(self):
		yield from self._locl
		yield from super().__dir__()

	def __getattr__(self, k):
		if k in self._locl:
			return self._locl[k]
		elif self._need_save:
			self.save()
			return getattr(self._module, k)
		elif self._module:
			return getattr(self._module, k)
		else:
			raise AttributeError("No attribute ", k)

	def __setattr__(self, k, v):
		if not callable(v):
			super().__setattr__(k, v)
			return
		self._set_source(k, getsource(v), func=v)

	def _set_source(self, k: str, source: str, func: callable = None):
		if func is None:
			holder = {}
			exec(source, holder)
			if k not in holder:
				raise NameError(
					"Function in source has a different name", k, source
				)
			func = holder[k]
		outdented = dedent_source(source)
		expr = Expr(parse(outdented))
		expr.value.body[0].name = k
		if k in self._ast_idx:
			self._ast.body[self._ast_idx[k]] = expr
		else:
			self._ast_idx[k] = len(self._ast.body)
			self._ast.body.append(expr)
		if self._filename is not None:
			self._need_save = True
		if isinstance(self._module, str):
			func.__module__ = self._module
		self._locl[k] = func
		self.send(self, attr=k, val=func)

	def __call__(self, v):
		if isinstance(self._module, str):
			v.__module__ = self._module
		elif hasattr(self._module, "__name__"):
			v.__module__ = self._module.__name__
		setattr(self, v.__name__, v)
		return v

	def __delattr__(self, k):
		del self._locl[k]
		del self._ast.body[self._ast_idx[k]]
		del self._ast_idx[k]
		for name in list(self._ast_idx):
			if name > k:
				self._ast_idx[name] -= 1
		if self._filename is not None:
			self._need_save = True
		self.send(self, attr=k, val=None)

	def save(self, reimport=True):
		if self._filename is None:
			return
		with open(self._filename, "w", encoding="utf-8") as outf:
			outf.write("# encoding: utf-8")
			Unparser(self._ast, outf)
		self._need_save = False
		if reimport:
			self.reimport()

	def reimport(self):
		if self._filename is None:
			return
		path, filename = os.path.split(self._filename)
		modname = filename[:-3]
		if modname in sys.modules:
			del sys.modules[modname]
		modname = filename[:-3]
		spec = importlib.util.spec_from_file_location(modname, self._filename)
		self._module = importlib.util.module_from_spec(spec)
		sys.modules[modname] = self._module
		spec.loader.exec_module(self._module)
		self._ast = parse(self._module.__loader__.get_data(self._filename))
		self._ast_idx = {}
		for i, node in enumerate(self._ast.body):
			if hasattr(node, "name"):
				self._ast_idx[node.name] = i
			elif hasattr(node, "__name__"):
				self._ast_idx[node.__name__] = i
		self.send(self, attr=None, val=None)

	def iterplain(self):
		for name, idx in self._ast_idx.items():
			yield name, unparse(self._ast.body[idx])

	def store_source(self, v, name=None):
		self._need_save = True
		outdented = dedent_source(v)
		mod = parse(outdented)
		expr = Expr(mod)
		if len(expr.value.body) != 1:
			raise ValueError("Tried to store more than one function")
		if name is None:
			name = expr.value.body[0].name
		else:
			expr.value.body[0].name = name
		if name in self._ast_idx:
			self._ast.body[self._ast_idx[name]] = expr
		else:
			self._ast_idx[name] = len(self._ast.body)
			self._ast.body.append(expr)
		locl = {}
		exec(compile(mod, self._filename, "exec"), {}, locl)
		self._locl.update(locl)
		self.send(self, attr=name, val=locl[name])

	def get_source(self, name):
		if name == "truth":
			return "def truth(*args):\n\treturn True"
		return unparse(self._ast.body[self._ast_idx[name]])

	@staticmethod
	def truth(*args):
		return True


class UniversalMapping(MutableMapping, Signal):
	"""Mapping for variables that are global but which I keep history for"""

	__slots__ = ["engine"]

	def __init__(self, engine):
		"""Store the engine and initialize my private dictionary of
		listeners.

		"""
		super().__init__()
		self.engine = engine

	def __iter__(self):
		return self.engine._universal_cache.iter_keys(*self.engine._btt())

	def __len__(self):
		return self.engine._universal_cache.count_keys(*self.engine._btt())

	def __getitem__(self, k):
		"""Get the current value of this key"""
		return wrapval(
			self,
			k,
			self._get_cache_now(k),
		)

	def _get_cache_now(self, k):
		return self.engine._universal_cache.retrieve(k, *self.engine._btt())

	def __setitem__(self, k, v):
		"""Set k=v at the current branch and tick"""
		branch, turn, tick = self.engine._nbtt()
		self.engine._universal_cache.store(k, branch, turn, tick, v)
		self.engine.query.universal_set(k, branch, turn, tick, v)
		self.send(self, key=k, val=v)

	def _set_cache_now(self, k, v):
		self.engine._universal_cache.store(k, *self.engine._btt(), v)

	def __delitem__(self, k):
		"""Unset this key for the present (branch, tick)"""
		branch, turn, tick = self.engine._nbtt()
		self.engine._universal_cache.store(k, branch, turn, tick, ...)
		self.engine.query.universal_del(k, branch, turn, tick)
		self.send(self, key=k, val=...)


class CharacterMapping(GraphsMapping, Signal):
	"""A mapping by which to access :class:`Character` objects.

	If a character already exists, you can always get its name here to
	get the :class:`Character` object. Deleting an item here will
	delete the character from the world, even if there are still
	:class:`Character` objects referring to it; those won't do
	anything useful anymore.

	"""

	engine = getatt("orm")

	def __init__(self, orm):
		GraphsMapping.__init__(self, orm)
		Signal.__init__(self)

	def __getitem__(self, name: Key | CharName) -> "Character":
		"""Return the named character, if it's been created.

		Try to use the cache if possible.

		"""
		from .character import Character

		name = CharName(name)
		if name not in self:
			raise KeyError("No such character", name)
		cache = self.engine._graph_objs
		if name not in cache:
			cache[name] = Character(
				self.engine, name, init_rulebooks=name not in self
			)
		ret = cache[name]
		if not isinstance(ret, Character):
			raise TypeError(
				"You put something weird in the Character cache", type(ret)
			)
		return ret

	def __setitem__(self, name: CharName, value: dict | nx.Graph):
		"""Make a new character by the given name, and initialize its data to
		the given value.

		"""
		self.engine._init_graph(name, "DiGraph", value)
		self.send(self, key=name, val=self.engine.character[name])

	def __delitem__(self, name: CharName):
		self.engine.del_character(name)
		self.send(self, key=name, val=None)


class CompositeDict(MutableMapping, Signal):
	"""Combine two dictionaries into one"""

	def __init__(self, d1, d2):
		"""Store dictionaries"""
		super().__init__()
		self.d1 = d1
		self.d2 = d2

	def __iter__(self):
		"""Iterate over both dictionaries' keys"""
		for k in self.d1:
			yield k
		for k in self.d2:
			yield k

	def __len__(self):
		"""Sum the lengths of both dictionaries"""
		return len(self.d1) + len(self.d2)

	def __contains__(self, item):
		return item in self.d1 or item in self.d2

	def __getitem__(self, k):
		"""Get an item from ``d1`` if possible, then ``d2``"""
		try:
			return self.d1[k]
		except KeyError:
			return self.d2[k]

	def __setitem__(self, key, value):
		self.d1[key] = value
		self.send(self, key=key, value=value)

	def __delitem__(self, key):
		deleted = False
		if key in self.d2:
			deleted = True
			del self.d2[key]
		if key in self.d1:
			deleted = True
			del self.d1[key]
		if not deleted:
			raise KeyError("{} is in neither of my wrapped dicts".format(key))
		self.send(self, key=key, value=None)

	def patch(self, d):
		"""Recursive update"""
		for k, v in d.items():
			if k in self:
				self[k].update(v)
			else:
				self[k] = deepcopy(v)



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import networkx as nx
import pytest

from lisien.exc import AmbiguousLeaderError


@pytest.fixture(scope="function")
def someplace(engy):
	yield engy.new_character("physical").new_place("someplace")


def test_contents(someplace):
	stuff = [someplace.new_thing(i) for i in range(10)]
	assert len(someplace.content) == len(stuff)
	assert len(someplace.contents()) == len(stuff)
	for i in range(10):
		assert i in someplace.content
		assert someplace.content[i] == stuff[i]
	for that in stuff:
		assert that in someplace.contents()
	for that in someplace.contents():
		assert that in stuff
	elsewhere = someplace.character.new_place("elsewhere")
	fust = elsewhere.new_thing(11)
	assert fust not in someplace.contents()
	assert 11 not in someplace.content
	with pytest.raises(KeyError):
		someplace.content[11]


def test_portal(someplace):
	assert not someplace.portal
	assert "there" not in someplace.portal
	there = someplace.character.new_place("there")
	assert not there.preportal
	assert "someplace" not in there.preportal
	someplace.character.new_portal("someplace", "there")
	assert someplace.portal
	assert "there" in someplace.portal
	assert "there" not in someplace.preportal
	assert there.preportal
	assert "someplace" in there.preportal
	assert "someplace" not in there.portal
	someplace.character.remove_edge("someplace", "there")
	assert "there" not in someplace.portal
	assert "someplace" not in there.preportal


def test_user(someplace):
	with pytest.raises(AmbiguousLeaderError):
		someplace.leader.only
	someone = someplace.engine.new_character("someone")
	someone.add_unit(someplace)
	assert someplace.leader.only is someone
	assert "someone" in someplace.leader
	assert someplace.leader["someone"] is someone
	noone = someplace.engine.new_character("noone")
	assert "noone" not in someplace.leader
	noone.add_unit(someplace)
	with pytest.raises(AmbiguousLeaderError):
		someplace.leader.only
	assert "noone" in someplace.leader
	assert someplace.leader["noone"] is noone


def test_rulebook(someplace):
	assert someplace.rulebook.name == ("physical", "someplace")
	someplace.rulebook = "imadeitup"
	assert someplace.rulebook.name == "imadeitup"


def test_deletion_after_keyframe(engine):
	phys = engine.new_character("physical", data=nx.grid_2d_graph(8, 8))
	del phys.place[5, 5]
	assert (5, 5) not in phys.place
	assert (5, 5) not in list(phys.place)
	while engine.turn < 20:
		engine.next_turn()
	assert (5, 5) not in phys.place
	assert (5, 5) not in list(phys.place)


def test_clear(engy):
	phys = engy.new_character("physical")
	place = phys.new_place("here")
	place["a"] = 1
	place["b"] = 2
	assert place["a"] == 1
	assert place["b"] == 2
	place.clear()
	assert "a" not in place
	with pytest.raises(KeyError):
		print(place["b"])



# This file is part of Lisien, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import logging
import threading
from queue import Queue

import cherrypy

from ..handle import EngineHandle


class LiSEHandleWebService(object):
	exposed = True

	def __init__(self, *args, **kwargs):
		if "logger" in kwargs:
			self.logger = kwargs["logger"]
		else:
			self.logger = kwargs["logger"] = logging.getLogger(__name__)
		self.cmdq = kwargs["cmdq"] = Queue()
		self.outq = kwargs["outq"] = Queue()
		self._handle_thread = threading.Thread(
			target=self._run_handle_forever,
			args=args,
			kwargs=kwargs,
			daemon=True,
		)
		self._handle_thread.start()

	@staticmethod
	def _run_handle_forever(*args, **kwargs):
		cmdq = kwargs.pop("cmdq")
		outq = kwargs.pop("outq")
		logger = kwargs.pop("logger")
		setup = kwargs.pop("setup", None)
		logq = Queue()

		def log(typ, data):
			if typ == "command":
				(cmd, args) = data
				logger.debug(
					"lisien thread {}: calling {}{}".format(
						threading.get_ident(), cmd, tuple(args)
					)
				)
			else:
				logger.debug(
					"lisien thread {}: returning {} (of type {})".format(
						threading.get_ident(), data, repr(type(data))
					)
				)

		def get_log_forever(logq):
			(level, data) = logq.get()
			getattr(logger, level)(data)

		engine_handle = EngineHandle(*args, logq=logq, **kwargs)
		if setup:
			setup(engine_handle._real)
		handle_log_thread = threading.Thread(
			target=get_log_forever, args=(logq,), daemon=True
		)
		handle_log_thread.start()
		while True:
			inst = cmdq.get()
			if inst == "shutdown":
				handle_log_thread.join()
				cmdq.close()
				outq.close()
				return 0
			cmd = inst.pop("command")
			silent = inst.pop("silent", False)
			log("command", (cmd, args))
			response = getattr(engine_handle, cmd)(**inst)
			if silent:
				continue
			log("result", response)
			outq.put(engine_handle._real.listify(response))

	@cherrypy.tools.accept(media="application/json")
	@cherrypy.tools.json_out()
	def GET(self):
		return cherrypy.session["LiSE_response"]

	@cherrypy.tools.json_out()
	def POST(self, **kwargs):
		silent = kwargs.get("silent", False)
		self.cmdq.put(kwargs)
		if silent:
			return None
		response = self.outq.get()
		cherrypy.session["LiSE_response"] = response
		return response

	def PUT(self, silent=False, **kwargs):
		silent = silent
		self.cmdq.put(kwargs)
		if not silent:
			cherrypy.session["LiSE_response"] = self.outq.get()

	def DELETE(self):
		cherrypy.session.pop("LiSE_response", None)
